/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Sequence/Matrix Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "seq.h"
#include "memalloc.h"
#include "mathx.h"

/******************************************************************************\
* Constructors and destructors.
\******************************************************************************/

matrix_t *seq2d_create(int xstart, int ystart, int xend, int yend)
{
	matrix_t *matrix;
	assert(xstart <= xend && ystart <= yend);
	if (!(matrix = matrix_create(yend - ystart, xend - xstart))) {
		return 0;
	}
	matrix->xstart_ = xstart;
	matrix->ystart_ = ystart;
	matrix->xend_ = xend;
	matrix->yend_ = yend;
	return matrix;
}

matrix_t *matrix_create(int numrows, int numcols)
{
	matrix_t *matrix;
	int i;

	if (!(matrix = malloc(sizeof(matrix_t)))) {
		return 0;
	}
	matrix->flags_ = 0;
	matrix->numrows_ = numrows;
	matrix->numcols_ = numcols;
	matrix->rows_ = 0;
	matrix->maxrows_ = numrows;
	matrix->data_ = 0;
	matrix->datasize_ = numrows * numcols;

	if (matrix->maxrows_ > 0) {
		if (!(matrix->rows_ = malloc(matrix->maxrows_ *
		  sizeof(fix_t *)))) {
			matrix_destroy(matrix);
			return 0;
		}
	}

	if (matrix->datasize_ > 0) {
		if (!(matrix->data_ = malloc(matrix->datasize_ *
		  sizeof(fix_t)))) {
			matrix_destroy(matrix);
			return 0;
		}
	}

	for (i = 0; i < numrows; ++i) {
		matrix->rows_[i] = &matrix->data_[i * matrix->numcols_];
	}

	for (i = 0; i < matrix->datasize_; ++i) {
		matrix->data_[i] = 0;
	}

	matrix->xstart_ = 0;
	matrix->ystart_ = 0;
	matrix->xend_ = matrix->numcols_;
	matrix->yend_ = matrix->numrows_;

	return matrix;
}

void matrix_destroy(matrix_t *matrix)
{
	if (matrix->data_) {
		assert(!(matrix->flags_ & MATRIX_REF));
		free(matrix->data_);
		matrix->data_ = 0;
	}
	if (matrix->rows_) {
		free(matrix->rows_);
		matrix->rows_ = 0;
	}
	free(matrix);
}

seq2d_t *seq2d_copy(seq2d_t *x)
{
	matrix_t *y;
	int i;
	int j;
	y = seq2d_create(seq2d_xstart(x), seq2d_ystart(x), seq2d_xend(x),
	  seq2d_yend(x));
	assert(y);
	for (i = 0; i < x->numrows_; ++i) {
		for (j = 0; j < x->numcols_; ++j) {
			*matrix_getref(y, i, j) = matrix_get(x, i, j);
		}
	}
	return y;
}

matrix_t *matrix_copy(matrix_t *x)
{
	matrix_t *y;
	int i;
	int j;
	y = matrix_create(x->numrows_, x->numcols_);
	for (i = 0; i < x->numrows_; ++i) {
		for (j = 0; j < x->numcols_; ++j) {
			*matrix_getref(y, i, j) = matrix_get(x, i, j);
		}
	}
	return y;
}

/******************************************************************************\
* Bind operations.
\******************************************************************************/

void seq2d_bindsub(matrix_t *s, matrix_t *s1, int xstart, int ystart,
  int xend, int yend)
{
	matrix_bindsub(s, s1, ystart - s1->ystart_, xstart - s1->xstart_,
	  yend - s1->ystart_ - 1, xend - s1->xstart_ - 1);
}

void matrix_bindsub(matrix_t *mat0, matrix_t *mat1, int r0, int c0,
  int r1, int c1)
{
	int i;

	mat0->flags_ |= MATRIX_REF;
	if (mat0->data_) {
		if (!(mat0->flags_ & MATRIX_REF)) {
			free(mat0->data_);
		}
		mat0->data_ = 0;
		mat0->datasize_ = 0;
	}
	if (mat0->rows_) {
		free(mat0->rows_);
		mat0->rows_ = 0;
	}
	mat0->numrows_ = r1 - r0 + 1;
	mat0->numcols_ = c1 - c0 + 1;
	mat0->maxrows_ = mat0->numrows_;
	mat0->rows_ = malloc(mat0->maxrows_ * sizeof(fix_t *));
	for (i = 0; i < mat0->numrows_; ++i) {
		mat0->rows_[i] = mat1->rows_[r0 + i] + c0;
	}

	mat0->xstart_ = mat1->xstart_ + c0;
	mat0->ystart_ = mat1->ystart_ + r0;
	mat0->xend_ = mat0->xstart_ + mat0->numcols_;
	mat0->yend_ = mat0->ystart_ + mat0->numrows_;
}

/******************************************************************************\
* Arithmetic operations.
\******************************************************************************/

int matrix_cmp(matrix_t *mat0, matrix_t *mat1)
{
	int i;
	int j;

	if (mat0->numrows_ != mat1->numrows_ || mat0->numcols_ !=
	  mat1->numcols_) {
		return 1;
	}
	for (i = 0; i < mat0->numrows_; i++) {
		for (j = 0; j < mat0->numcols_; j++) {
			if (matrix_get(mat0, i, j) != matrix_get(mat1, i, j)) {
				return 1;
			}
		}
	}
	return 0;
}

void matrix_mulpow2(matrix_t *matrix, int n)
{
	int i;
	int j;
	fix_t *rowstart;
	int rowstep;
	fix_t *data;

	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			*data = (*data >= 0) ? ((*data) << n) :
			  (-((-(*data)) << n));
		}
	}
}

void matrix_divpow2(matrix_t *matrix, int n)
{
	int i;
	int j;
	fix_t *rowstart;
	int rowstep;
	fix_t *data;

	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			*data = (*data >= 0) ? ((*data) >> n) :
			  (-((-(*data)) >> n));
		}
	}
}

void matrix_clip(matrix_t *matrix, fix_t minval, fix_t maxval)
{
	int i;
	int j;
	fix_t v;
	fix_t *rowstart;
	fix_t *data;
	int rowstep;

	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		data = rowstart;
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			v = *data;
			if (fix_lt(v, minval)) {
				*data = minval;
			} else if (fix_gt(v, maxval)) {
				*data = maxval;
			}
		}
	}
}

void matrix_asr(matrix_t *matrix, int n)
{
	int i;
	int j;
	fix_t *rowstart;
	int rowstep;
	fix_t *data;

	assert(n >= 0);
	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			*data >>= n;
		}
	}
}

void matrix_asl(matrix_t *matrix, int n)
{
	int i;
	int j;
	fix_t *rowstart;
	int rowstep;
	fix_t *data;

	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			*data <<= n;
		}
	}
}

void matrix_andeq(matrix_t *matrix, fix_t val)
{
	int i;
	int j;
	int rowstep;
	fix_t *rowstart;
	fix_t *data;

	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		data = rowstart;
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			*data &= val;
		}
	}
}

/******************************************************************************\
* Code.
\******************************************************************************/

int matrix_resize(matrix_t *matrix, int numrows, int numcols)
{
	int size;
	int i;

	size = numrows * numcols;
	if (size > matrix->datasize_ || numrows > matrix->maxrows_) {
		return -1;
	}

	matrix->numrows_ = numrows;
	matrix->numcols_ = numcols;

	for (i = 0; i < numrows; ++i) {
		matrix->rows_[i] = &matrix->data_[numcols * i];
	}

	return 0;
}

static int outval(stream_t *out, fix_t val)
{
	unsigned int x;
	int s;
	int i;

	if (val < 0) {
		s = 1;
		x = -val;
	} else {
		s = 0;
		x = val;
	}

	if (stream_putc(out, ((x >> 24) & 0x7f) | (s << 7)) == EOF) {
		return -1;
	}
	for (i = 0; i < 3; ++i) {
		x <<= 8;
		if (stream_putc(out, (x >> 24) & 0xff) == EOF) {
			return -1;
		}
	}

	return 0;
}

int matrix_outputraw(matrix_t *matrix, stream_t *out)
{
	int i;
	int j;
	fix_t v;

	outval(out, matrix_numrows(matrix));
	outval(out, matrix_numcols(matrix));
	for (i = 0; i < matrix_numrows(matrix); ++i) {
		for (j = 0; j < matrix_numcols(matrix); ++j) {
			v = matrix_get(matrix, i, j);
			outval(out, v);
		}
	}

	return 0;
}

static int inval(stream_t *in, fix_t *val)
{
	int v;
	unsigned int x;
	int s;
	int c;
	int i;

	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	s = c >> 7;
	x = c & 0x7f;
	for (i = 0; i < 3; ++i) {
		if ((c = stream_getc(in)) == EOF) {
			return -1;
		}
		x = (x << 8) | c;
	}

	v = x;
	if (s) {
		v = -v;
	}

	*val = v;
	return 0;
}

matrix_t *matrix_inputraw(stream_t *in)
{
	int numrows;
	int numcols;
	matrix_t *matrix;
	int i;
	int j;
	fix_t v;

	if (inval(in, &v)) {
		return 0;
	}
	numrows = v;
	if (inval(in, &v)) {
		return 0;
	}
	numcols = v;
	matrix = matrix_create(numrows, numcols);
	for (i = 0; i < matrix_numrows(matrix); ++i) {
		for (j = 0; j < matrix_numcols(matrix); ++j) {
			if (inval(in, &v)) {
				matrix_destroy(matrix);
				return 0;
			}
			matrix_set(matrix, i, j, v);
		}
	}

	return matrix;
}

int MATRIX_OUTPUT(matrix_t *matrix, FILE *out)
{
	int i;
	int j;
	fix_t x;

	fprintf(out, "%d %d\n", matrix_numrows(matrix), matrix_numcols(matrix));
	for (i = 0; i < matrix_numrows(matrix); ++i) {
		for (j = 0; j < matrix_numcols(matrix); ++j) {
			x = (matrix_get(matrix, i, j));
			fprintf(out, "%d", x);
			if (j < matrix_numcols(matrix) - 1) {
				fprintf(out, " ");
			}
		}
		fprintf(out, "\n");
	}
	return 0;
}

int matrix_output(matrix_t *matrix, FILE *out)
{
	int i;
	int j;
	double x;

	fprintf(out, "%d %d\n", matrix_numrows(matrix), matrix_numcols(matrix));
	for (i = 0; i < matrix_numrows(matrix); ++i) {
		for (j = 0; j < matrix_numcols(matrix); ++j) {
			x = fixtodbl(matrix_get(matrix, i, j));
			fprintf(out, "%f", x);
			if (j < matrix_numcols(matrix) - 1) {
				fprintf(out, " ");
			}
		}
		fprintf(out, "\n");
	}
	return 0;
}

void matrix_setall(matrix_t *matrix, fix_t val)
{
	int i;
	int j;
	fix_t *rowstart;
	int rowstep;
	fix_t *data;

	rowstep = matrix_rowstep(matrix);
	for (i = matrix->numrows_, rowstart = matrix->rows_[0]; i > 0; --i,
	  rowstart += rowstep) {
		for (j = matrix->numcols_, data = rowstart; j > 0; --j,
		  ++data) {
			*data = val;
		}
	}
}

matrix_t *matrix_input(FILE *in)
{
	matrix_t *matrix;
	int i;
	int j;
	double x;
	int numrows;
	int numcols;

	if (fscanf(in, "%d %d", &numrows, &numcols) != 2)
		return 0;
	if (!(matrix = matrix_create(numrows, numcols)))
		return 0;

	/* Get matrix data. */
	for (i = 0; i < matrix_numrows(matrix); i++) {
		for (j = 0; j < matrix_numcols(matrix); j++) {
			if (fscanf(in, "%lf", &x) != 1) {
				matrix_destroy(matrix);
				return 0;
			}
			matrix_set(matrix, i, j, dbltofix(x));
		}
	}

	return matrix;
}

int matrix_output2(matrix_t *matrix, FILE *out)
{
	int i;
	int j;
	int x;
char buf[256];

	fprintf(out, "%d %d\n", matrix_numrows(matrix), matrix_numcols(matrix));
	for (i = 0; i < matrix_numrows(matrix); ++i) {
		for (j = 0; j < matrix_numcols(matrix); ++j) {
			x = matrix_get(matrix, i, j);
			sprintf(buf, "%c%x", ((x < 0) ? '-' : '+'), ((x < 0) ? (-x) : x));
			fprintf(out, "%-4s ", buf);
			if (j < matrix_numcols(matrix) - 1) {
				fprintf(out, " ");
			}
		}
		fprintf(out, "\n");
	}
	return 0;
}









seq_t *seq_upsample(seq_t *x, int m)
{
	seq_t *z;
	int i;

	if (!(z = seq_create(seq_start(x) * m, (seq_end(x) - 1) * m + 1)))
		return 0;
	for (i = seq_start(z); i < seq_end(z); i++) {
		*seq_getref(z, i) = (!mod(i, m)) ? seq_get(x, i / m) :
		  inttofix(0);
	}

	return z;
}

fix_t seq_norm(seq_t *x)
{
	fix_t s;
	int i;

	s = inttofix(0);
	for (i = seq_start(x); i < seq_end(x); i++) {
		s = fix_add(s, fix_mul(seq_get(x, i), seq_get(x, i)));
	}

	return dbltofix(sqrt(fixtodbl(s)));
}

seq_t *seq_conv(seq_t *x, seq_t *y)
{
	int i;
	int j;
	int k;
	seq_t *z;
	fix_t s;
	fix_t v;

	z = seq_create(seq_start(x) + seq_start(y),
	  seq_end(x) + seq_end(y) - 1);
	assert(z);
	for (i = seq_start(z); i < seq_end(z); i++) {
		s = inttofix(0);
		for (j = seq_start(y); j < seq_end(y); j++) {
			k = i - j;
			if (k < seq_start(x) || k >= seq_end(x)) {
				v = FIX_ZERO;
			} else {
				v = seq_get(x, k);
			}
			s = fix_add(s, fix_mul(seq_get(y, j), v));
		}
		*seq_getref(z, i) = s;
	}

	return z;
}
