/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * I/O Stream Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "types.h"
#include <assert.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif
#if defined(WIN32) || defined(HAVE_IO_H)
#include <io.h>
#endif
#include "stream.h"
#include "memalloc.h"

/******************************************************************************\
* Local function prototypes.
\******************************************************************************/

static int strtoopenmode(const char *s);
static void stream_destroy(stream_t *stream);
static stream_t *stream_create(void);
static void stream_initbuf(stream_t *stream, int bufmode, char *buf,
  int bufsize);

static int mem_read(stream_obj_t *obj, char *buf, int cnt);
static int mem_write(stream_obj_t *obj, char *buf, int cnt);
static long mem_seek(stream_obj_t *obj, long offset, int origin);
static int mem_close(stream_obj_t *obj);

static int file_read(stream_obj_t *obj, char *buf, int cnt);
static int file_write(stream_obj_t *obj, char *buf, int cnt);
static long file_seek(stream_obj_t *obj, long offset, int origin);
static int file_close(stream_obj_t *obj);

/******************************************************************************\
* Local data.
\******************************************************************************/

static stream_ops_t stream_fileops = {
	file_read,
	file_write,
	file_seek,
	file_close
};

static stream_ops_t stream_memops = {
	mem_read,
	mem_write,
	mem_seek,
	mem_close
};

/******************************************************************************\
* Code for opening and closing streams.
\******************************************************************************/

static stream_t *stream_create()
{
	stream_t *stream;

	if (!(stream = malloc(sizeof(stream_t)))) {
		return 0;
	}
	stream->openmode_ = 0;
	stream->bufmode_ = 0;
	stream->flags_ = 0;
	stream->bufbase_ = 0;
	stream->bufstart_ = 0;
	stream->bufsize_ = 0;
	stream->ptr_ = 0;
	stream->cnt_ = 0;
	stream->ops_ = 0;
	stream->obj_ = 0;
	stream->rwcnt_ = 0;
	stream->rwlimit_ = -1;

	return stream;
}

stream_t *stream_memopen(char *buf, int bufsize)
{
	stream_t *stream;
	stream_memobj_t *obj;

	if (!(stream = stream_create())) {
		return 0;
	}

	/* A stream associated with a memory buffer is always opened
	for both reading and writing in binary mode. */
	stream->openmode_ = STREAM_READ | STREAM_WRITE | STREAM_BINARY;

	/* Since the stream data is already resident in memory, buffering
	is not necessary. */
	/* But... It still may be faster to use buffering anyways. */
	stream_initbuf(stream, STREAM_FULLBUF, 0, 0);

	/* Select the operations for a memory stream. */
	stream->ops_ = &stream_memops;

	/* Allocate memory for the underlying memory stream object. */
	if (!(obj = malloc(sizeof(stream_memobj_t)))) {
		stream_destroy(stream);
		return 0;
	}
	stream->obj_ = (void *) obj;

	/* Initialize a few important members of the memory stream object. */
	obj->myalloc_ = 0;
	obj->buf_ = 0;

	/* If the buffer size specified is nonpositive, then the buffer
	is allocated internally and automatically grown as needed. */
	if (bufsize <= 0) {
		obj->bufsize_ = 1024;
		obj->growable_ = 1;
	} else {
		obj->bufsize_ = bufsize;
		obj->growable_ = 0;
	}
	if (buf) {
		obj->buf_ = (unsigned char *) buf;
	} else {
		obj->buf_ = malloc(obj->bufsize_ * sizeof(char));
		obj->myalloc_ = 1;
	}
	if (!obj->buf_) {
		stream_close(stream);
		return 0;
	}

	if (bufsize > 0 && buf) {
		/* If a buffer was supplied by the caller and its length is positive,
		  make the associated buffer data appear in the stream initially. */
		obj->len_ = bufsize;
	} else {
		/* The stream is initially empty. */
		obj->len_ = 0;
	}
	obj->pos_ = 0;
	
	return stream;
}

stream_t *stream_fopen(const char *filename, const char *mode)
{
	stream_t *stream;
	int *obj;
	int openflags;

	/* Allocate a stream object. */
	if (!(stream = stream_create())) {
		return 0;
	}

	/* Parse the mode string. */
	stream->openmode_ = strtoopenmode(mode);

	/* Determine the correct flags to use for opening the file. */
	if ((stream->openmode_ & STREAM_READ) &&
	  (stream->openmode_ & STREAM_WRITE)) {
		openflags = O_RDWR;
	} else if (stream->openmode_ & STREAM_READ) {
		openflags = O_RDONLY;
	} else if (stream->openmode_ & STREAM_WRITE) {
		openflags = O_WRONLY;
	} else {
		openflags = 0;
	}
	if (stream->openmode_ & STREAM_APPEND) {
		openflags |= O_APPEND;
	}
	if (stream->openmode_ & STREAM_BINARY) {
		openflags |= O_BINARY;
	}
	if (stream->openmode_ & STREAM_CREATE) {
		openflags |= O_CREAT | O_TRUNC;
	}

	/* Allocate space for the underlying file stream object. */
	if (!(obj = malloc(sizeof(int)))) {
		stream_destroy(stream);
		return 0;
	}
	stream->obj_ = (void *) obj;

	/* Select the operations for a file stream object. */
	stream->ops_ = &stream_fileops;

	/* Open the underlying file. */
	if ((*obj = open(filename, openflags, STREAM_PERMS)) < 0) {
		stream_destroy(stream);
		return 0;
	}

	/* By default, use full buffering for this type of stream. */
	stream_initbuf(stream, STREAM_FULLBUF, 0, 0);

	return stream;
}

stream_t *stream_tmpfile()
{
	stream_t *stream;
	int *obj;
	char filename[L_tmpnam + 1];

	if (!(stream = stream_create())) {
		return 0;
	}

	/* A temporary file stream is always opened for both reading and
	writing in binary mode. */
	stream->openmode_ = STREAM_READ | STREAM_WRITE | STREAM_BINARY;

	/* Allocate memory for the underlying temporary file object. */
	if (!(obj = malloc(sizeof(int)))) {
		stream_destroy(stream);
		return 0;
	}
	stream->obj_ = obj;

	/* Choose a file name. */
	tmpnam(filename);

	/* Open the underlying file. */
	if ((*obj = open(filename, O_CREAT | O_EXCL | O_RDWR | O_TRUNC | O_BINARY,
	  STREAM_PERMS)) < 0) {
		stream_destroy(stream);
		return 0;
	}

	/* Unlink the file so that it will disappear if the program
	terminates abnormally. */
	/* Under UNIX, one can unlink an open file and continue to do I/O
	on it.  This also appears to work under MS Windows.  If you use
	some other operating system, this may not work. :-)  If this
    functionality is not supported by the operating system, the unlink
	operation must be deferred until the file is closed. */
	unlink(filename);

	/* Use full buffering. */
	stream_initbuf(stream, STREAM_FULLBUF, 0, 0);

	stream->ops_ = &stream_fileops;

	return stream;
}

stream_t *stream_fdopen(int fd, const char *mode)
{
	stream_t *stream;
	int *obj;

	/* Allocate a stream object. */
	if (!(stream = stream_create())) {
		return 0;
	}

	/* Parse the mode string. */
	stream->openmode_ = strtoopenmode(mode);

#if defined(WIN32)
	/* Argh!!!  Someone ought to banish text mode (i.e., O_TEXT) to the
	  greatest depths of purgatory! */
	/* Ensure that the file descriptor is in binary mode, if the caller
	  has specified the binary mode flag.  Arguably, the caller ought to
	  take care of this, but text mode is a ugly wart anyways, so we save
	  the caller some grief by handling this within the stream library. */
	/* This ugliness is mainly for the benefit of those who run the
	  JasPer software under Windows from shells that insist on opening
	  files in text mode.  For example, in the Cygwin environment,
	  shells often open files in text mode when I/O redirection is
	  used.  Grr... */
	if (stream->openmode_ & STREAM_BINARY) {
		setmode(fd, O_BINARY);
	}
#endif

	/* Allocate space for the underlying file stream object. */
	if (!(obj = malloc(sizeof(int)))) {
		stream_destroy(stream);
		return 0;
	}
	stream->obj_ = (void *) obj;
	*obj = fd;

	/* By default, use full buffering for this type of stream. */
	stream_initbuf(stream, STREAM_FULLBUF, 0, 0);

	/* Select the operations for a file stream object. */
	stream->ops_ = &stream_fileops;

	return stream;
}

static void stream_destroy(stream_t *stream)
{
	/* If the memory for the buffer was allocated with malloc, free
	this memory. */
	if ((stream->bufmode_ & STREAM_FREEBUF) && stream->bufbase_) {
		free(stream->bufbase_);
		stream->bufbase_ = 0;
	}
	free(stream);
}

int stream_close(stream_t *stream)
{
	/* Flush buffer if necessary. */
	stream_flush(stream);

	/* Close the underlying stream object. */
	(*stream->ops_->close_)(stream->obj_);

	stream_destroy(stream);

	return 0;
}

/******************************************************************************\
* Code for reading and writing streams.
\******************************************************************************/

int stream_getc_func(stream_t *stream)
{
	assert(stream->ptr_ - stream->bufbase_ <= stream->bufsize_ +
	  STREAM_MAXPUTBACK);
	return stream_getc_macro(stream);
}

int stream_putc_func(stream_t *stream, int c)
{
	assert(stream->ptr_ - stream->bufstart_ <= stream->bufsize_);
	return stream_putc_macro(stream, c);
}

int stream_ungetc(stream_t *stream, int c)
{
	if (!stream->ptr_ || stream->ptr_ == stream->bufbase_) {
		return -1;
	}
	--stream->rwcnt_;
	--stream->ptr_;
	++stream->cnt_;
	*stream->ptr_ = c;
	return 0;
}

int stream_read(stream_t *stream, void *buf, int cnt)
{
	int n;
	int c;
	char *bufptr;

	bufptr = buf;

	n = 0;
	while (n < cnt) {
		if ((c = stream_getc(stream)) == EOF) {
			return n;
		}
		*bufptr++ = c;
		++n;
	}

	return n;
}

int stream_write(stream_t *stream, const void *buf, int cnt)
{
	int n;
	const char *bufptr;

	bufptr = buf;

	n = 0;
	while (n < cnt) {
		if (stream_putc(stream, *bufptr) == EOF) {
			return n;
		}
		++bufptr;
		++n;
	}

	return n;
}

/* Note: This function uses a fixed size buffer.  Therefore, it cannot
  handle invocations that will produce more output than can be held
  by the buffer. */
int stream_printf(stream_t *stream, const char *fmt, ...)
{
	va_list ap;
	char buf[4096];
	int ret;

	va_start(ap, fmt);
	ret = vsprintf(buf, fmt, ap);
	stream_puts(stream, buf);
	va_end(ap);
	return ret;
}

int stream_puts(stream_t *stream, const char *s)
{
	while (*s != '\0') {
		if (stream_putc_macro(stream, *s) == EOF) {
			return -1;
		}
		++s;
	}
	return 0;
}

char *stream_gets(stream_t *stream, char *buf, int bufsize)
{
	int c;
	char *bufptr;
	assert(bufsize > 0);

	bufptr = buf;
	while (bufsize > 1) {
		if ((c = stream_getc(stream)) == EOF) {
			break;
		}
		*bufptr++ = c;
		--bufsize;
		if (c == '\n') {
			break;
		}
	}
	*bufptr = '\0';
	return buf;
}

int stream_gobble(stream_t *stream, int n)
{
	int m;
	m = n;
	for (m = n; m > 0; --m) {
		if (stream_getc(stream) == EOF) {
			return n - m;
		}
	}
	return n;
}

/******************************************************************************\
* Code for getting and setting the stream position.
\******************************************************************************/

int stream_isseekable(stream_t *stream)
{
	if (stream->ops_ == &stream_memops) {
		return 1;
	} else if (stream->ops_ == &stream_fileops) {
		if ((*stream->ops_->seek_)(stream->obj_, 0, SEEK_CUR) < 0) {
			return 0;
		}
		return 1;
	} else {
		return 0;
	}
}

int stream_rewind(stream_t *stream)
{
	return stream_seek(stream, 0, SEEK_SET);
}

long stream_seek(stream_t *stream, long offset, int origin)
{
	long newpos;

	/* The buffer cannot be in use for both reading and writing. */
	assert(!((stream->bufmode_ & STREAM_RDBUF) && (stream->bufmode_ &
	  STREAM_WRBUF)));

	/* Reset the EOF indicator (since we may not be at the EOF anymore). */
	stream->flags_ &= ~STREAM_EOF;

	if (stream->bufmode_ & STREAM_RDBUF) {
		if (origin == SEEK_CUR) {
			offset -= stream->cnt_;
		}
	} else if (stream->bufmode_ & STREAM_WRBUF) {
		if (stream_flush(stream)) {
			return -1;
		}
	}
	stream->cnt_ = 0;
	stream->ptr_ = stream->bufstart_;
	stream->bufmode_ &= ~(STREAM_RDBUF | STREAM_WRBUF);

	if ((newpos = (*stream->ops_->seek_)(stream->obj_, offset, origin))
	  < 0) {
		return -1;
	}

	return newpos;
}

long stream_tell(stream_t *stream)
{
	int adjust;
	int offset;

	if (stream->bufmode_ & STREAM_RDBUF) {
		adjust = -stream->cnt_;
	} else if (stream->bufmode_ & STREAM_WRBUF) {
		adjust = stream->ptr_ - stream->bufstart_;
	} else {
		adjust = 0;
	}

	if ((offset = (*stream->ops_->seek_)(stream->obj_, 0, SEEK_CUR)) < 0) {
		return -1;
	}

	return offset + adjust;
}

/******************************************************************************\
* Buffer initialization code.
\******************************************************************************/

static void stream_initbuf(stream_t *stream, int bufmode, char *buf,
  int bufsize)
{
	/* If this function is being called, the buffer should not have been
	  initialized yet. */
	assert(!stream->bufbase_);

	if (bufmode != STREAM_UNBUF) {
		/* The full- or line-buffered mode is being employed. */
		if (!buf) {
			/* The caller has not specified a buffer to employ, so allocate
			  one. */
			if ((stream->bufbase_ = malloc(STREAM_BUFSIZE +
			  STREAM_MAXPUTBACK))) {
				stream->bufmode_ |= STREAM_FREEBUF;
				stream->bufsize_ = STREAM_BUFSIZE;
			} else {
				/* The buffer allocation has failed.  Resort to unbuffered
				  operation. */
				stream->bufbase_ = stream->tinybuf_;
				stream->bufsize_ = 1;
			}
		} else {
			/* The caller has specified a buffer to employ. */
			/* The buffer must be large enough to accommodate maximum
			  putback. */
			assert(bufsize > STREAM_MAXPUTBACK);
			stream->bufbase_ = CAST(uchar *, buf);
			stream->bufsize_ = bufsize - STREAM_MAXPUTBACK;
		}
	} else {
		/* The unbuffered mode is being employed. */
		/* A buffer should not have been supplied by the caller. */
		assert(!buf);
		/* Use a trivial one-character buffer. */
		stream->bufbase_ = stream->tinybuf_;
		stream->bufsize_ = 1;
	}
	stream->bufstart_ = &stream->bufbase_[STREAM_MAXPUTBACK];
	stream->ptr_ = stream->bufstart_;
	stream->cnt_ = 0;
	stream->bufmode_ |= bufmode & STREAM_BUFMODEMASK;
}

/******************************************************************************\
* Buffer filling and flushing code.
\******************************************************************************/

int stream_flush(stream_t *stream)
{
	if (stream->bufmode_ & STREAM_RDBUF) {
		return 0;
	}
	return stream_flushbuf(stream, EOF);
}

int stream_fillbuf(stream_t *stream, int getflag)
{
	int c;

	/* The stream must not be in an error or EOF state. */
	if ((stream->flags_ & (STREAM_ERRMASK)) != 0) {
		return EOF;
	}

	/* The stream must be open for reading. */
	if ((stream->openmode_ & STREAM_READ) == 0) {
		return EOF;
	}

	/* Make a half-hearted attempt to confirm that the buffer is not
	currently being used for writing.  This check is not intended
	to be foolproof! */
	assert((stream->bufmode_ & STREAM_WRBUF) == 0);

	assert(stream->ptr_ - stream->bufstart_ <= stream->bufsize_);

	/* Mark the buffer as being used for reading. */
	stream->bufmode_ |= STREAM_RDBUF;

	/* Read new data into the buffer. */
	stream->ptr_ = stream->bufstart_;
	if ((stream->cnt_ = (*stream->ops_->read_)(stream->obj_,
	  (char *) stream->bufstart_, stream->bufsize_)) <= 0) {
		if (stream->cnt_ < 0) {
			stream->flags_ |= STREAM_ERR;
		} else {
			stream->flags_ |= STREAM_EOF;
		}
		stream->cnt_ = 0;
		return EOF;
	}

	assert(stream->cnt_ > 0);
	/* Get or peek at the first character in the buffer. */
	c = (getflag) ? stream_getc2(stream) : (*stream->ptr_);

	return c;
}

int stream_flushbuf(stream_t *stream, int c)
{
	int len;
	int n;

	/* The stream should not be in an error or EOF state. */
	if ((stream->flags_ & (STREAM_ERRMASK)) != 0) {
		return EOF;
	}

	/* The stream must be open for writing. */
	if ((stream->openmode_ & (STREAM_WRITE | STREAM_APPEND)) == 0) {
		return EOF;
	}

	/* The buffer should not currently be in use for reading. */
	assert(!(stream->bufmode_ & STREAM_RDBUF));

	/* Note: Do not use the quantity stream->cnt to determine the number
	of characters in the buffer!  Depending on how this function was
	called, the stream->cnt value may be "off-by-one". */
	len = stream->ptr_ - stream->bufstart_;
	if (len > 0) {
		n = (*stream->ops_->write_)(stream->obj_, (char *)
		  stream->bufstart_, len);
		if (n != len) {
			stream->flags_ |= STREAM_ERR;
			return EOF;
		}
	}
	stream->cnt_ = stream->bufsize_;
	stream->ptr_ = stream->bufstart_;

	stream->bufmode_ |= STREAM_WRBUF;

	if (c != EOF) {
		assert(stream->cnt_ > 0);
		stream_putc2(stream, c);
	}

	return 0;
}

/******************************************************************************\
* Miscellaneous code.
\******************************************************************************/

static int strtoopenmode(const char *s)
{
	int openmode = 0;
	while (*s != '\0') {
		switch (*s) {
		case 'r':
			openmode |= STREAM_READ;
			break;
		case 'w':
			openmode |= STREAM_WRITE | STREAM_CREATE;
			break;
		case 'b':
			openmode |= STREAM_BINARY;
			break;
		case 'a':
			openmode |= STREAM_APPEND;
			break;
		case '+':
			openmode |= STREAM_READ | STREAM_WRITE;
			break;
		default:
			break;
		}
		++s;
	}
	return openmode;
}

int stream_copy(stream_t *out, stream_t *in, int n)
{
	int all;
	int c;
	int m;

	all = (n < 0) ? 1 : 0;

	m = n;
	while (all || m > 0) {
		if ((c = stream_getc_macro(in)) == EOF) {
			/* The next character of input could not be read. */
			/* Return with an error if an I/O error occured
			  (not including EOF) or if an explicit copy count
			  was specified. */
			return (!all || stream_error(in)) ? (-1) : 0;
		}
		if (stream_putc_macro(out, c) == EOF) {
			return -1;
		}
		--m;
	}
	return 0;
}

long stream_setrwcount(stream_t *stream, long rwcnt)
{
	int old;

	old = stream->rwcnt_;
	stream->rwcnt_ = rwcnt;
	return old;
}

int stream_display(stream_t *stream, FILE *fp, int n)
{
	unsigned char buf[16];
	int i;
	int j;
	int m;
	int c;
	int display;
	int cnt;

	cnt = n - (n % 16);
	display = 1;

	for (i = 0; i < n; i += 16) {
		if (n > 16 && i > 0) {
			display = (i >= cnt) ? 1 : 0;
		}
		if (display) {
			fprintf(fp, "%08x:", i);
		}
		m = min(n - i, 16);
		for (j = 0; j < m; ++j) {
			if ((c = stream_getc(stream)) == EOF) {
				abort();
				return -1;
			}
			buf[j] = c;
		}
		if (display) {
			for (j = 0; j < m; ++j) {
				fprintf(fp, " %02x", buf[j]);
			}
			fputc(' ', fp);
			for (; j < 16; ++j) {
				fprintf(fp, "   ");
			}
			for (j = 0; j < m; ++j) {
				if (isprint(buf[j])) {
					fputc(buf[j], fp);
				} else {
					fputc(' ', fp);
				}
			}
			fprintf(fp, "\n");
		}


	}
	return 0;
}

long stream_length(stream_t *stream)
{
	long oldpos;
	long pos;
	if ((oldpos = stream_tell(stream)) < 0) {
		return -1;
	}
	if (stream_seek(stream, 0, SEEK_END) < 0) {
		return -1;
	}
	if ((pos = stream_tell(stream)) < 0) {
		return -1;
	}
	if (stream_seek(stream, oldpos, SEEK_SET) < 0) {
		return -1;
	}
	return pos;
}

/******************************************************************************\
* Memory stream object.
\******************************************************************************/

static int mem_read(stream_obj_t *obj, char *buf, int cnt)
{
	int n;
	stream_memobj_t *m = (stream_memobj_t *)obj;
	n = m->len_ - m->pos_;
	cnt = min(n, cnt);
	memcpy(buf, &m->buf_[m->pos_], cnt);
	m->pos_ += cnt;
	return cnt;
}

static int mem_resize(stream_memobj_t *m, int bufsize)
{
	unsigned char *buf;

	assert(m->buf_);
	if (!(buf = realloc(m->buf_, bufsize * sizeof(unsigned char)))) {
		return -1;
	}
	m->buf_ = buf;
	m->bufsize_ = bufsize;
	return 0;
}

static int mem_write(stream_obj_t *obj, char *buf, int cnt)
{
	int n;
	int ret;
	stream_memobj_t *m = (stream_memobj_t *)obj;
	long newbufsize;
	long newpos;

	newpos = m->pos_ + cnt;
	if (newpos > m->bufsize_ && m->growable_) {
		newbufsize = m->bufsize_;
		while (newbufsize < newpos) {
			newbufsize <<= 1;
			assert(newbufsize >= 0);
		}
		if (mem_resize(m, newbufsize)) {
			return -1;
		}
	}
	if (m->pos_ > m->len_) {
		/* The current position is beyond the end of the file, so
		  pad the file to the current position with zeros. */
		n = min(m->pos_, m->bufsize_) - m->len_;
		if (n > 0) {
			memset(&m->buf_[m->len_], 0, n);
			m->len_ += n;
		}
		if (m->pos_ != m->len_) {
			/* The buffer is not big enough. */
			return 0;
		}
	}
	n = m->bufsize_ - m->pos_;
	ret = min(n, cnt);
	if (ret > 0) {
		memcpy(&m->buf_[m->pos_], buf, ret);
		m->pos_ += ret;
	}
	if (m->pos_ > m->len_) {
		m->len_ = m->pos_;
	}
assert(ret == cnt);
	return ret;
}

static long mem_seek(stream_obj_t *obj, long offset, int origin)
{
	stream_memobj_t *m = (stream_memobj_t *)obj;
	long newpos;

	switch (origin) {
	case SEEK_SET:
		newpos = offset;
		break;
	case SEEK_END:
		newpos = m->len_ - offset;
		break;
	case SEEK_CUR:
		newpos = m->pos_ + offset;
		break;
	default:
		abort();
		break;
	}
	if (newpos < 0) {
		return -1;
	}
	m->pos_ = newpos;

	return m->pos_;
}

static int mem_close(stream_obj_t *obj)
{
	stream_memobj_t *m = (stream_memobj_t *)obj;
	if (m->myalloc_ && m->buf_) {
		free(m->buf_);
		m->buf_ = 0;
	}
	free(obj);
	return 0;
}

/******************************************************************************\
* File stream object.
\******************************************************************************/

static int file_read(stream_obj_t *obj, char *buf, int cnt)
{
	int fd;
	fd = *((int *)obj);
	return read(fd, buf, cnt);
}

static int file_write(stream_obj_t *obj, char *buf, int cnt)
{
	int fd;
	fd = *((int *)obj);
	return write(fd, buf, cnt);
}

static long file_seek(stream_obj_t *obj, long offset, int origin)
{
	int fd;
	fd = *((int *)obj);
	return lseek(fd, offset, origin);
}

static int file_close(stream_obj_t *obj)
{
	int fd;
	fd = *((int *)obj);
	free(obj);
	return close(fd);
}
