/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * I/O Stream Library
 *
 * $Id$
 */

#ifndef STREAM_H
#define STREAM_H

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif
#include "types.h"

/******************************************************************************\
* Macros.
\******************************************************************************/

/* On most UNIX systems, we probably need to define O_BINARY ourselves. */
#ifndef O_BINARY
#define O_BINARY	0
#endif

#ifndef min
#define min(a, b)	(((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#define max(a, b)	(((a) > (b)) ? (a) : (b))
#endif

/******************************************************************************\
* Constants.
\******************************************************************************/

/*
 * Stream open flags.
 */

#define STREAM_READ	0x0001
/* The stream was opened for reading. */
#define STREAM_WRITE	0x0002
/* The stream was opened for writing. */
#define STREAM_APPEND	0x0004
/* The stream was opened for appending. */
#define STREAM_BINARY	0x0008
/* The stream was opened in binary mode. */
#define STREAM_CREATE	0x0010
/* The stream should be created/truncated. */

/*
 * Stream buffering flags.
 */

#define STREAM_UNBUF	0x0000
/* The stream is unbuffered. */
#define STREAM_LINEBUF	0x0001
/* The stream is line buffered. */
#define STREAM_FULLBUF	0x0002
/* The stream is fully buffered. */
#define	STREAM_BUFMODEMASK	0x000f

#define STREAM_FREEBUF	0x0008
/* The memory associated with the buffer needs to be deallocated when the
  stream is destroyed. */
#define STREAM_RDBUF	0x0010
/* The buffer is currently being used for reading. */
#define STREAM_WRBUF	0x0020
/* The buffer is currently being used for writing. */

/*
 * Stream error flags.
 */

#define STREAM_EOF	0x0001
/* The end-of-file has been encountered (on reading). */
#define STREAM_ERR	0x0002
/* An I/O error has been encountered on the stream. */
#define	STREAM_RWLIMIT	0x0004
/* The read/write limit has been exceeded. */
#define STREAM_ERRMASK	(STREAM_EOF | STREAM_ERR | STREAM_RWLIMIT)

/*
 * Other miscellaneous constants.
 */

#define STREAM_BUFSIZE	8192
/* The default buffer size (for fully-buffered operation). */
#define STREAM_PERMS	0666
/* The default permission mask for file creation. */

#define	STREAM_MAXPUTBACK	16
/* The maximum number of characters that can always be put back on a stream. */

/******************************************************************************\
* Types.
\******************************************************************************/

/*
 * Generic file object.
 */

typedef void stream_obj_t;

/*
 * Generic file object operations.
 */

typedef struct {

	int (*read_)(stream_obj_t *obj, char *buf, int cnt);
	/* Read characters from a file object. */

	int (*write_)(stream_obj_t *obj, char *buf, int cnt);
	/* Write characters to a file object. */

	long (*seek_)(stream_obj_t *obj, long offset, int origin);
	/* Set the position for a file object. */

	int (*close_)(stream_obj_t *obj);
	/* Close a file object. */

} stream_ops_t;

/*
 * Stream object.
 */

typedef struct {

	int openmode_;
	/* The mode in which the stream was opened. */

	int bufmode_;
	/* The buffering mode. */

	int flags_;
	/* The stream status. */

	uchar *bufbase_;
	/* The start of the buffer area to use for reading/writing. */

	uchar *bufstart_;
	/* The start of the buffer area excluding the extra initial space for
	  character putback. */

	int bufsize_;
	/* The buffer size. */

	uchar *ptr_;
	/* The current position in the buffer. */

	int cnt_;
	/* The number of characters that must be read/written before
	the buffer needs to be filled/flushed. */

	uchar tinybuf_[STREAM_MAXPUTBACK + 1];
	/* A trivial buffer to be used for unbuffered operation. */

	stream_ops_t *ops_;
	/* The operations for the underlying stream file object. */

	stream_obj_t *obj_;
	/* The underlying stream file object. */

	long rwcnt_;
	/* The number of characters read/written. */

	long rwlimit_;
	/* The maximum number of characters that may be read/written. */

} stream_t;

/*
 * Regular file object.
 */

typedef int stream_fileobj_t;
/* This is simply a file descriptor. */

/*
 * Memory file object.
 */

typedef struct {

	uchar *buf_;
	/* The data associated with this file. */

	int bufsize_;
	/* The allocated size of the buffer for holding file data. */

	int_fast32_t len_;
	/* The length of the file. */

	int_fast32_t pos_;
	/* The seek position. */

	int growable_;
	/* Is the buffer growable? */

	int myalloc_;
	/* Was the buffer allocated internally? */

} stream_memobj_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

/* Open a file as a stream. */
stream_t *stream_fopen(const char *filename, const char *mode);

/* Open a memory buffer as a stream. */
stream_t *stream_memopen(char *buf, int bufsize);

/* Open a file descriptor as a stream. */
stream_t *stream_fdopen(int fd, const char *mode);

/* Open a temporary file as a stream. */
stream_t *stream_tmpfile();

/* Close a stream. */
int stream_close(stream_t *stream);

/* Get the EOF indicator for a stream. */
#define stream_eof(stream) \
	(((stream)->flags_ & STREAM_EOF) != 0)

/* Get the error indicator for a stream. */
#define stream_error(stream) \
	(((stream)->flags_ & STREAM_ERR) != 0)

/* Clear the error indicator for a stream. */
#define stream_clearerr(stream) \
	((stream)->flags_ &= ~(STREAM_ERR | STREAM_EOF))

/* Read a character from a stream. */
#define stream_getc_macro(stream) \
	((!((stream)->flags_ & (STREAM_ERR | STREAM_EOF | STREAM_RWLIMIT))) ? \
	(((stream)->rwlimit_ >= 0 && (stream)->rwcnt_ >= (stream)->rwlimit_) ? \
	  (stream->flags_ |= STREAM_RWLIMIT, EOF) : stream_getc2(stream)) : EOF)

/* Write a character to a stream. */
#define stream_putc_macro(stream, c) \
	((!((stream)->flags_ & (STREAM_ERR | STREAM_EOF | STREAM_RWLIMIT))) ? \
	(((stream)->rwlimit_ >= 0 && (stream)->rwcnt_ >= (stream)->rwlimit_) ? \
	  (stream->flags_ |= STREAM_RWLIMIT, EOF) : stream_putc2(stream, c)) : EOF)

/* Look at the next character to be read from a stream without actually
  removing it from the stream. */
#define	stream_peekc(stream) \
	(((stream)->cnt_ <= 0) ? stream_fillbuf(stream, 0) : \
	  ((int)(*(stream)->ptr_)))

/* Read characters from a stream into a buffer. */
int stream_read(stream_t *stream, void *buf, int cnt);

/* Write characters from a buffer to a stream. */
int stream_write(stream_t *stream, const void *buf, int cnt);

/* Set the current position within the stream. */
long stream_seek(stream_t *stream, long offset, int origin);

/* Flush any pending output to a stream. */
int stream_flush(stream_t *stream);

/* Write formatted output to a stream. */
int stream_printf(stream_t *stream, const char *fmt, ...);

/* Write a string to a stream. */
int stream_puts(stream_t *stream, const char *s);

/* Read a line of input from a stream. */
char *stream_gets(stream_t *stream, char *buf, int bufsize);

/* Seek to the beginning of a stream. */
int stream_rewind(stream_t *stream);

/* Get the current position within the stream. */
long stream_tell(stream_t *stream);

/* Read a character from a stream. */
#ifdef DEBUG
#define	stream_getc(stream)	stream_getc_func(stream)
#else
#define stream_getc(stream)	stream_getc_macro(stream)
#endif

/* Write a character to a stream. */
#ifdef DEBUG
#define stream_putc(stream, c)	stream_putc_func(stream, c)
#else
#define stream_putc(stream, c)	stream_putc_macro(stream, c)
#endif

/* Copy data from one stream to another. */
int stream_copy(stream_t *dst, stream_t *src, int n);

/* Is it possible to seek on this stream? */
int stream_isseekable(stream_t *stream);

/* Get the read/write limit for a stream. */
#define	stream_getrwlimit(stream) \
	(((const stream_t *)(stream))->rwlimit_)

/* Set the read/write limit for a stream. */
int stream_setrwlimit(stream_t *stream, long rwlimit);

/* Get the read/write count for a stream. */
#define	stream_getrwcount(stream) \
	(((const stream_t *)(stream))->rwcnt_)

/* Set the read/write count for a stream. */
long stream_setrwcount(stream_t *stream, long rwcnt);

int stream_ungetc(stream_t *stream, int c);
/* Put a character back on a stream. */

int stream_display(stream_t *stream, FILE *fp, int n);

int stream_gobble(stream_t *stream, int n);

long stream_length(stream_t *stream);

/******************************************************************************\
* Internal functions.
\******************************************************************************/

/* The following functions are for internal use only!  If you call them
directly, you will die a horrible, miserable, and painful death! */

/* These prototypes need to be here for the sake of the stream_getc and
stream_putc macros. */
int stream_fillbuf(stream_t *stream, int getflag);
int stream_flushbuf(stream_t *stream, int c);

int stream_getc_func(stream_t *stream);
int stream_putc_func(stream_t *stream, int c);

#define stream_getc2(stream) \
	((--(stream)->cnt_ < 0) ? stream_fillbuf(stream, 1) : \
	  (++(stream)->rwcnt_, (int)(*(stream)->ptr_++)))

#define stream_putc2(stream, c) \
	(((stream)->bufmode_ |= STREAM_WRBUF, --(stream)->cnt_ < 0) ? \
	  stream_flushbuf((stream), (uchar)(c)) : \
	  (++(stream)->rwcnt_, (int)(*(stream)->ptr_++ = (c))))

#ifdef __cplusplus
}
#endif

#endif
