/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Bit Stream Library
 *
 * $Id$
 */

#ifndef BS_H
#define BS_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include "stream.h"
#include "memalloc.h"

/******************************************************************************\
* Constants and types.
\******************************************************************************/

#define	BITSTREAM_READ	0x01
/* Bit stream open for reading. */
#define	BITSTREAM_WRITE	0x02
/* Bit stream open for writing. */

#define	BITSTREAM_NOCLOSE	0x01
/* Do not close underlying character stream. */

#define	BITSTREAM_EOF	0x02
#define	BITSTREAM_ERR	0x04

/* Bit stream. */

typedef struct {

	int flags_;
	/* Some miscellaneous flags. */

	uint_fast16_t buf_;
	/* The input/output buffer. */

	int c_;
	/* A temporary variable used by the get/put macros (i.e., a hack). */

	int cnt_;
	/* The number of bits remaining in the byte being read/written. */

	stream_t *stream_;
	/* The underlying stream associated with this bit stream. */

	int openmode_;
	/* The mode in which this bit stream was opened. */

} bitstream_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

bitstream_t *bitstream_fopen(char *path, char *mode);
/* Open a file as a bit stream. */

bitstream_t *bitstream_sopen(stream_t *stream, char *mode);
/* Open a stream as a bit stream. */

int bitstream_close(bitstream_t *bitstream);
/* Close a bit stream. */

#if defined(DEBUG)
#define	bitstream_getbit(bitstream) \
	bitstream_getbit_func(bitstream)
#else
#define bitstream_getbit(bitstream) \
	bitstream_getbit_macro(bitstream)
#endif
/* Read a bit from a bit stream. */

#if defined(DEBUG)
#define	bitstream_putbit(bitstream, v) \
	bitstream_putbit_func(bitstream, v)
#else
#define	bitstream_putbit(bitstream, v) \
	bitstream_putbit_macro(bitstream, v)
#endif
/* Write a bit to a bit stream. */

long bitstream_getbits(bitstream_t *bitstream, int n);
/* Read one or more bits from a bit stream. */

int bitstream_putbits(bitstream_t *bitstream, int n, long v);
/* Write one or more bits to a bit stream. */

int bitstream_align(bitstream_t *bitstream);
/* Align the current position within the bit stream to the next byte
  boundary. */

int bitstream_inalign(bitstream_t *bitstream, int fillmask, int filldata);
/* Align the current position in the bit stream with the next byte boundary,
  ensuring that certain bits consumed in the process match a particular
  pattern. */

int bitstream_outalign(bitstream_t *bitstream, int filldata);
/* Align the current position in the bit stream with the next byte boundary,
  writing bits from the specified pattern (if necessary) in the process. */

int bitstream_needalign(bitstream_t *bitstream);
/* Check if a bit stream needs alignment. */

int bitstream_pending(bitstream_t *bitstream);

int bitstream_setdbglevel(int dbglevel);
/* Set the debug level for the bit stream library. */

#define bitstream_eof(bitstream) \
	((bitstream)->flags_ & BITSTREAM_EOF)

/******************************************************************************\
* Internals.
\******************************************************************************/

/* DO NOT DIRECTLY INVOKE ANY OF THE MACROS OR FUNCTIONS BELOW.  THEY ARE
  FOR INTERNAL USE ONLY. */

int bitstream_getbit_func(bitstream_t *bitstream);

int bitstream_putbit_func(bitstream_t *bitstream, int v);

int bitstream_fillbuf(bitstream_t *bitstream);

#define	bitstream_getbit_macro(bitstream) \
	(assert((bitstream)->openmode_ & BITSTREAM_READ), \
	  (--(bitstream)->cnt_ >= 0) ? \
	  (((bitstream)->buf_ >> (bitstream)->cnt_) & 1) : \
	  bitstream_fillbuf(bitstream))

#define bitstream_putbit_macro(bitstream, bit) \
	(assert((bitstream)->openmode_ & BITSTREAM_WRITE), \
	  (--(bitstream)->cnt_ < 0) ? \
	  ((bitstream)->buf_ = ((bitstream)->buf_ << 8) & 0xffff, \
	  (bitstream)->cnt_ = ((bitstream)->buf_ == 0xff00) ? 6 : 7, \
	  (bitstream)->buf_ |= ((bit) & 1) << (bitstream)->cnt_, \
	  (stream_putc((bitstream)->stream_, (bitstream)->buf_ >> 8) == EOF) \
	  ? (EOF) : ((bit) & 1)) : \
	  ((bitstream)->buf_ |= ((bit) & 1) << (bitstream)->cnt_, \
	  (bit) & 1))

#endif
