/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * $Id$
 */

#ifndef JPCENC_H
#define JPCENC_H

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "t2cod.h"
#include "mqenc.h"
#include "jpccod.h"
#include "tagtree.h"
#include "seq.h"
#include "flt.h"
#include "cs.h"

/******************************************************************************\
* Constants.
\******************************************************************************/

#define	NUMEXTRABITS	NMSEDEC_FRACBITS
#define	BADRDSLOPE	(-1)

/******************************************************************************\
* Coding parameters types.
\******************************************************************************/

/* Component coding paramters. */
typedef struct {
	uint_fast8_t sampgrdstepx;
	uint_fast8_t sampgrdstepy;
	uint_fast8_t sampgrdsubstepx;
	uint_fast8_t sampgrdsubstepy;
	uint_fast8_t prec;
	bool sgnd;
	uint_fast16_t numstepsizes;
	uint_fast16_t stepsizes[JPC_MAXBANDS];
} enc_ccp_t;

/* Tile coding parameters. */
typedef struct {

	bool intmode;
	/* The coding mode. */

	uint_fast8_t csty;
	/* The coding style (i.e., SOP, EPH). */

	uint_fast8_t prg;
	/* The progression order. */

	uint_fast8_t mctid;
	/* The multicomponent transform. */

	uint_fast16_t numlyrs;
	/* The number of layers. */

	fix_t *ilyrrates;
	/* The normalized bit rates associated with the various
	  intermediate layers. */

} enc_tcp_t;

/* Tile-component coding parameters. */
typedef struct {

	uint_fast8_t csty;
	/* The coding style (i.e., explicit precinct sizes). */

	uint_fast8_t maxrlvls;
	/* The maximum number of resolution levels allowed. */

	uint_fast16_t cblkwidthexpn;
	/* The nominal code block width. */

	uint_fast16_t cblkheightexpn;
	/* The nominal code block height. */

	uint_fast8_t cblksty;
	/* The code block style parameters (e.g., lazy, terminate all,
	  segmentation symbols, causal, reset probability models). */

	uint_fast8_t qmfbid;

	uint_fast16_t prcwidthexpns[JPC_MAXRLVLS];
	/* The precinct width values. */

	uint_fast16_t prcheightexpns[JPC_MAXRLVLS];
	/* The precinct height values. */

	uint_fast8_t numgbits;
	/* The number of guard bits. */

} enc_tccp_t;

typedef struct {

	int debug;
	/* The debug level. */

	uint_fast32_t imgareatlx;
	/* The horizontal offset from the origin of the reference grid to the
	  left edge of the image area. */

	uint_fast32_t imgareatly;
	/* The vertical offset from the origin of the reference grid to the
	  top edge of the image area. */

	uint_fast32_t refgrdwidth;
	/* The horizontal offset from the origin of the reference grid to the
	  right edge of the image area (plus one). */

	uint_fast32_t refgrdheight;
	/* The vertical offset from the origin of the reference grid to the
	  bottom edge of the image area (plus one). */

	uint_fast32_t tilegrdoffx;
	/* The horizontal offset from the origin of the tile grid to the
	  origin of the reference grid. */

	uint_fast32_t tilegrdoffy;
	/* The vertical offset from the origin of the tile grid to the
	  origin of the reference grid. */

	uint_fast32_t tilewidth;
	/* The nominal tile width in units of the image reference grid. */

	uint_fast32_t tileheight;
	/* The nominal tile height in units of the image reference grid. */

	uint_fast32_t numhtiles;
	/* The number of tiles spanning the image area in the horizontal
	  direction. */

	uint_fast32_t numvtiles;
	/* The number of tiles spanning the image area in the vertical
	  direction. */

	uint_fast32_t numtiles;
	/* The number of tiles. */

	uint_fast16_t numcmpts;

	enc_ccp_t *ccps;

	enc_tcp_t tcp;

	enc_tccp_t tccp;

	uint_fast32_t totalsize;
	/* The target code stream length in bytes. */

	uint_fast32_t rawsize;
	/* The raw (i.e., uncompressed) size of the image in bytes. */

} enc_cp_t;

/******************************************************************************\
* Types.
\******************************************************************************/

struct enc_comp_s;
struct enc_lvl_s;

typedef struct {

	int start;
	/* The starting offset for this pass. */

	int end;
	/* The ending offset for this pass. */

	int type;
	/* The type of data in this pass. */

	int term;
	/* Flag indicating that this pass is terminated. */

	mqencstate_t mqencstate;
	/* The entropy coder state after coding this pass. */

	int lyrno;
	/* The layer to which this pass has been assigned. */

	flt_t rdslope;

	flt_t wmsedec;

	flt_t cumwmsedec;

	long nmsedec;

} enc_pass_t;

typedef struct {

	int numpasses;
	/* The number of passes. */

	enc_pass_t *passes;
	/* The per-pass information. */

	int numencpasses;
	/* The number of passes encoded so far. */

	int numimsbs;
	/* The number of insignificant MSBs. */

	int numlenbits;
	/* The number of bits used to encode pass data lengths. */

	stream_t *stream;
	/* The byte stream for this code block. */

	mqenc_t *mqenc;
	/* The entropy encoder. */

	matrix_t *data;
	/* The data for this code block. */

	matrix_t *flags;
	/* The state for this code block. */

	int numbps;

	enc_pass_t *curpass;
	/* The next pass to be encoded. */


	/* For rate control... */
	enc_pass_t *savedcurpass;
	int savednumlenbits;
	int savednumencpasses;

	struct enc_prc_s *prc;

} enc_cblk_t;

typedef struct enc_prc_s {

	uint_fast32_t tlx;

	uint_fast32_t tly;

	uint_fast32_t brx;

	uint_fast32_t bry;

	int numhcblks;
	/* The number of coding blocks spanning the band in the horizontal
	direction. */

	int numvcblks;
	/* The number of coding blocks spanning the band in the vertical
	direction. */

	int numcblks;
	/* The number of coding blocks associated with this band. */

	enc_cblk_t *cblks;
	/* The per-coding-block information. */

	tagtree_t *incltree;
	/* The inclusion tag tree. */

	tagtree_t *nlibtree;
	/* The insignifcant MSBs tag tree. */

	/* For rate control... */
	tagtree_t *savincltree;
	tagtree_t *savnlibtree;

struct enc_band_s *band;

} enc_prc_t;

typedef struct enc_band_s {

	enc_prc_t *prcs;
	/* The per precinct information. */

	matrix_t *data;
	/* The coefficient data for this band. */

	int orient;
	/* The orientation of this band (i.e., LL, LH, HL, or HH). */

	int numbps;
	/* The number of bit planes associated with this band. */

	fix_t absstepsize;
	int stepsize;

	fix_t synweight;
	/* The L2 norm of the synthesis basis functions associated with
	  this band.  (The MCT is not considered in this value.) */

	int analgain;

	struct enc_rlvl_s *rlvl;

} enc_band_t;

typedef struct enc_rlvl_s {

	uint_fast32_t tlx;
	/* The x-coordinate of the top-left corner of the tile-component
	  at this resolution. */

	uint_fast32_t tly;
	/* The y-coordinate of the top-left corner of the tile-component
	  at this resolution. */

	uint_fast32_t brx;
	/* The x-coordinate of the bottom-right corner of the tile-component
	  at this resolution (plus one). */

	uint_fast32_t bry;
	/* The y-coordinate of the bottom-right corner of the tile-component
	  at this resolution (plus one). */

	int prcwidthexpn;
	/* The exponent value for the nominal precinct width measured
	  relative to the associated LL band. */

	int prcheightexpn;
	/* The exponent value for the nominal precinct height measured
	  relative to the associated LL band. */

	int numhprcs;
	/* The number of precincts spanning the resolution level in the
	  horizontal direction. */

	int numvprcs;
	/* The number of precincts spanning the resolution level in the
	  vertical direction. */

	int numprcs;
	/* The number of precincts. */

	int cbgwidthexpn;
	/* The exponent value for the nominal code block group width.
	  This quantity is associated with the next lower resolution level
	  (assuming that there is one). */

	int cbgheightexpn;
	/* The exponent value for the nominal code block group height.
	  This quantity is associated with the next lower resolution level
	  (assuming that there is one). */

	uint_fast16_t cblkwidthexpn;
	/* The exponent value for the code block width. */

	uint_fast16_t cblkheightexpn;
	/* The exponent value for the code block height. */

	int numbands;
	/* The number of bands associated with this resolution level. */

	enc_band_t *bands;
	/* The per-band information. */

	struct enc_tcmpt_s *tcmpt;
	/* The parent tile-component. */

} enc_rlvl_t;

typedef struct enc_tcmpt_s {

	uint_fast16_t numrlvls;
	/* The number of resolution levels. */

	enc_rlvl_t *rlvls;
	/* The per-resolution-level information. */

	matrix_t *data;
	/* The tile-component data. */

	int qmfbid;
	int numbands;
	tsfb_t *tsfb;
	fix_t synweight;

	int prcwidthexpns[JPC_MAXRLVLS];
	int prcheightexpns[JPC_MAXRLVLS];
	int cblkwidthexpn;
	int cblkheightexpn;

	int csty;
	/* explicit precint sizes */

	int cblksty;
	/* Code block style. */

	uint_fast16_t numstepsizes;
	uint_fast16_t stepsizes[JPC_MAXBANDS];

	struct enc_tile_s *tile;
	/* The parent tile. */

} enc_tcmpt_t;

typedef struct enc_tile_s {

	uint_fast32_t tileno;

	uint_fast32_t tlx;
	/* The x-coordinate of the top-left corner of the tile measured with
	  respect to the reference grid. */

	uint_fast32_t tly;
	/* The y-coordinate of the top-left corner of the tile measured with
	  respect to the reference grid. */

	uint_fast32_t brx;
	/* The x-coordinate of the bottom-right corner of the tile measured
	  with respect to the reference grid (plus one). */

	uint_fast32_t bry;
	/* The y-coordinate of the bottom-right corner of the tile measured
	  with respect to the reference grid (plus one). */

	/* uint_fast16_t numparts; */
	/* The number of tile parts. */

	/* uint_fast16_t partno; */
	/* The current part number. */

	uint_fast8_t csty;
	/* The coding style. */

	uint_fast8_t prg;
	/* The progression order. */

	uint_fast16_t numlyrs;
	/* The number of layers. */

	uint_fast8_t mctid;
	/* The MCT to employ (if any). */

	pi_t *pi;
	/* The packet iterator (used to determine the order of packet
	  generation). */

	bool intmode;
	/* The coding mode (i.e., integer or real). */

	uint_fast32_t *lyrsizes;

	enc_tcmpt_t *tcmpts;
	/* The per tile-component information. */

	int numtcmpts;

	uint_fast32_t rawsize;
	/* The raw (i.e., uncompressed) size of this tile. */

} enc_tile_t;

typedef struct enc_s {

	image_t *image;
	/* The image being encoded. */

	stream_t *out;
	/* The output stream. */

	enc_cp_t *cp;
	/* The encoder options. */

	enc_tile_t *curtile;

	jpc_cstate_t *cstate;

	/* For future use... */
	/* roiinfo_t roiinfo; */

	uint_fast32_t len;
	/* The number of bytes output so far. */

	uint_fast32_t mainbodysize;

	jpc_ms_t *mrk;

	stream_t *tmpstream;
	/* For a tile. */

} enc_t;

/******************************************************************************\
* Functions.
\******************************************************************************/

void jpenc_setdbglvl(int lvl);

#endif
