/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Multicomponent Transform Code
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <assert.h>

#include "seq.h"
#include "memalloc.h"
#include "mct.h"

/******************************************************************************\
* Code.
\******************************************************************************/

/* Compute the forward RCT. */

void rct(matrix_t *c0, matrix_t *c1, matrix_t *c2)
{
	int numrows;
	int numcols;
	int i;
	int j;
	fix_t *c0p;
	fix_t *c1p;
	fix_t *c2p;

	numrows = matrix_numrows(c0);
	numcols = matrix_numcols(c0);

	/* All three matrices must have the same dimensions. */
	assert(matrix_numrows(c1) == numrows && matrix_numcols(c1) == numcols
	  && matrix_numrows(c2) == numrows && matrix_numcols(c2) == numcols);

	for (i = 0; i < numrows; i++) {
		c0p = matrix_getref(c0, i, 0);
		c1p = matrix_getref(c1, i, 0);
		c2p = matrix_getref(c2, i, 0);
		for (j = numcols; j > 0; --j) {
			int r;
			int g;
			int b;
			int y;
			int u;
			int v;
			r = *c0p;
			g = *c1p;
			b = *c2p;
			y = (r + (g << 1) + b) >> 2;
			u = b - g;
			v = r - g;
			*c0p++ = y;
			*c1p++ = u;
			*c2p++ = v;
		}
	}
}

/* Compute the inverse RCT. */

void irct(matrix_t *c0, matrix_t *c1, matrix_t *c2)
{
	int numrows;
	int numcols;
	int i;
	int j;
	fix_t *c0p;
	fix_t *c1p;
	fix_t *c2p;

	numrows = matrix_numrows(c0);
	numcols = matrix_numcols(c0);

	/* All three matrices must have the same dimensions. */
	assert(matrix_numrows(c1) == numrows && matrix_numcols(c1) == numcols
	  && matrix_numrows(c2) == numrows && matrix_numcols(c2) == numcols);

	for (i = 0; i < numrows; i++) {
		c0p = matrix_getref(c0, i, 0);
		c1p = matrix_getref(c1, i, 0);
		c2p = matrix_getref(c2, i, 0);
		for (j = numcols; j > 0; --j) {
			int r;
			int g;
			int b;
			int y;
			int u;
			int v;
			y = *c0p;
			u = *c1p;
			v = *c2p;
			g = y - ((u + v) >> 2);
			r = v + g;
			b = u + g;
			*c0p++ = r;
			*c1p++ = g;
			*c2p++ = b;
		}
	}
}

void ict(matrix_t *c0, matrix_t *c1, matrix_t *c2)
{
	int numrows;
	int numcols;
	int i;
	int j;
	fix_t r;
	fix_t g;
	fix_t b;
	fix_t y;
	fix_t u;
	fix_t v;
	fix_t *c0p;
	fix_t *c1p;
	fix_t *c2p;

	numrows = matrix_numrows(c0);
	assert(matrix_numrows(c1) == numrows && matrix_numrows(c2) == numrows);
	numcols = matrix_numcols(c0);
	assert(matrix_numcols(c1) == numcols && matrix_numcols(c2) == numcols);
	for (i = 0; i < numrows; ++i) {
		c0p = matrix_getref(c0, i, 0);
		c1p = matrix_getref(c1, i, 0);
		c2p = matrix_getref(c2, i, 0);
		for (j = numcols; j > 0; --j) {
			r = *c0p;
			g = *c1p;
			b = *c2p;
			y = fix_add3(fix_mul(dbltofix(0.299), r), fix_mul(dbltofix(0.587), g),
			  fix_mul(dbltofix(0.114), b));
			u = fix_add3(fix_mul(dbltofix(-0.16875), r), fix_mul(dbltofix(-0.33126), g),
			  fix_mul(dbltofix(0.5), b));
			v = fix_add3(fix_mul(dbltofix(0.5), r), fix_mul(dbltofix(-0.41869), g),
			  fix_mul(dbltofix(-0.08131), b));
			*c0p++ = y;
			*c1p++ = u;
			*c2p++ = v;
		}
	}
}

void iict(matrix_t *c0, matrix_t *c1, matrix_t *c2)
{
	int numrows;
	int numcols;
	int i;
	int j;
	fix_t r;
	fix_t g;
	fix_t b;
	fix_t y;
	fix_t u;
	fix_t v;
	fix_t *c0p;
	fix_t *c1p;
	fix_t *c2p;

	numrows = matrix_numrows(c0);
	assert(matrix_numrows(c1) == numrows && matrix_numrows(c2) == numrows);
	numcols = matrix_numcols(c0);
	assert(matrix_numcols(c1) == numcols && matrix_numcols(c2) == numcols);
	for (i = 0; i < numrows; ++i) {
		c0p = matrix_getref(c0, i, 0);
		c1p = matrix_getref(c1, i, 0);
		c2p = matrix_getref(c2, i, 0);
		for (j = numcols; j > 0; --j) {
			y = *c0p;
			u = *c1p;
			v = *c2p;
			r = fix_add(y, fix_mul(dbltofix(1.402), v));
			g = fix_add3(y, fix_mul(dbltofix(-0.34413), u),
			  fix_mul(dbltofix(-0.71414), v));
			b = fix_add(y, fix_mul(dbltofix(1.772), u));
			*c0p++ = r;
			*c1p++ = g;
			*c2p++ = b;
		}
	}
}


fix_t mct_getsynweight(int mctid, int cmptno)
{
	fix_t synweight;

	switch (mctid) {
	case MCT_RCT:
		switch (cmptno) {
		case 0:
			synweight = dbltofix(sqrt(3.0));
			break;
		case 1:
			synweight = dbltofix(sqrt(0.6875));
			break;
		case 2:
			synweight = dbltofix(sqrt(0.6875));
			break;
		}
		break;
	case MCT_ICT:
		switch (cmptno) {
		case 0:
			synweight = dbltofix(sqrt(3.0000));
			break;
		case 1:
			synweight = dbltofix(sqrt(3.2584));
			break;
		case 2:
			synweight = dbltofix(sqrt(2.4755));
			break;
		}
		break;
	default:
		synweight = FIX_ONE;
		break;
	}

	return synweight;
}
