/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Tier 1 Decoder
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "fix.h"
#include "bs.h"
#include "stream.h"
#include "mqdec.h"
#include "t1dec.h"
#include "t1cod.h"
#include "jpcdec.h"
#include "memalloc.h"
#include "util.h"

/******************************************************************************\
*
\******************************************************************************/

static int dec_decodecblk(jpd_t *dec, jpd_tile_t *tile, jpd_tcomp_t *tcomp, jpd_band_t *band,
  jpd_cblk_t *cblk, int dopartial, int maxlyrs);
static int dec_sigpass(jpd_t *dec, mqdec_t *mqdec, int bitpos, int orient,
  int vcausalflag, matrix_t *flags, matrix_t *data);
static int dec_rawsigpass(jpd_t *dec, bitstream_t *in, int bitpos,
  int vcausalflag, matrix_t *flags, matrix_t *data);
static int dec_refpass(jpd_t *dec, mqdec_t *mqdec, int bitpos, int vcausalflag,
  matrix_t *flags, matrix_t *data);
static int dec_rawrefpass(jpd_t *dec, bitstream_t *in, int bitpos,
  int vcausalflag, matrix_t *flags, matrix_t *data);
static int dec_clnpass(jpd_t *dec, mqdec_t *mqdec, int bitpos, int orient,
  int vcausalflag, int segsymflag, matrix_t *flags, matrix_t *data);

#if defined(DEBUG)
static long t1dec_cnt = 0;
#endif

#if !defined(DEBUG)
#define	T1D_GETBIT(mqdec, v, passtypename, symtypename) \
	((v) = mqdec_getbit(mqdec))
#else
#define	T1D_GETBIT(mqdec, v, passtypename, symtypename) \
{ \
	(v) = mqdec_getbit(mqdec); \
	if (getdbglevel() >= 100) { \
		fprintf(stderr, "index = %ld; passtype = %s; symtype = %s; sym = %d\n", t1dec_cnt, passtypename, symtypename, v); \
		++t1dec_cnt; \
	} \
}
#endif
#define	T1D_GETBITNOSKEW(mqdec, v, passtypename, symtypename) \
	T1D_GETBIT(mqdec, v, passtypename, symtypename)

#if !defined(DEBUG)
#define	T1D_RAWGETBIT(bitstream, v, passtypename, symtypename) \
	((v) = bitstream_getbit(bitstream))
#else
#define	T1D_RAWGETBIT(bitstream, v, passtypename, symtypename) \
{ \
	(v) = bitstream_getbit(bitstream); \
	if (getdbglevel() >= 100) { \
		fprintf(stderr, "index = %ld; passtype = %s; symtype = %s; sym = %d\n", t1dec_cnt, passtypename, symtypename, v); \
		++t1dec_cnt; \
	} \
}
#endif

/******************************************************************************\
* Code.
\******************************************************************************/

int dec_decodecblks(jpd_t *dec, jpd_tile_t *tile)
{
	jpd_tcomp_t *tcomp;
	int compcnt;
	jpd_rlvl_t *rlvl;
	int rlvlcnt;
	jpd_band_t *band;
	int bandcnt;
	jpd_prc_t *prc;
	int prccnt;
	jpd_cblk_t *cblk;
	int cblkcnt;

	for (compcnt = dec->numcomps, tcomp = tile->tcomps; compcnt > 0;
	  --compcnt, ++tcomp) {
		for (rlvlcnt = tcomp->numrlvls, rlvl = tcomp->rlvls;
		  rlvlcnt > 0; --rlvlcnt, ++rlvl) {
			if (!rlvl->bands) {
				continue;
			}
			for (bandcnt = rlvl->numbands, band = rlvl->bands;
			  bandcnt > 0; --bandcnt, ++band) {
				if (!band->data) {
					continue;
				}
				for (prccnt = rlvl->numprcs, prc = band->prcs;
				  prccnt > 0; --prccnt, ++prc) {
					if (!prc->cblks) {
						continue;
					}
					for (cblkcnt = prc->numcblks,
					  cblk = prc->cblks; cblkcnt > 0;
					  --cblkcnt, ++cblk) {
						if (dec_decodecblk(dec, tile, tcomp,
						  band, cblk, 1, JPC_MAXLYRS)) {
							return -1;
						}
					}
				}

			}
		}
	}

	return 0;
}

static int dec_decodecblk(jpd_t *dec, jpd_tile_t *tile, jpd_tcomp_t *tcomp, jpd_band_t *band,
  jpd_cblk_t *cblk, int dopartial, int maxlyrs)
{
	jpd_seg_t *seg;
	int i;
	int bpno;
	int passtype;
	int ret;
	int compno;
	int filldata;
	int fillmask;

	compno = tcomp - tile->tcomps;

	if (!cblk->flags) {
		/* Note: matrix is assumed to be zeroed */
		if (!(cblk->flags = matrix_create(matrix_numrows(cblk->data) +
		  2, matrix_numcols(cblk->data) + 2))) {
			return -1;
		}
	}

	seg = cblk->segs.head;
	while (seg && (seg != cblk->curseg || dopartial) && (maxlyrs < 0 ||
	  seg->lyrno < maxlyrs)) {
		assert(seg->numpasses >= seg->maxpasses || dopartial);
		assert(seg->stream);
		stream_rewind(seg->stream);
		stream_setrwcount(seg->stream, 0);
		if (seg->type == SEG_MQ) {
			if (!cblk->mqdec) {
				if (!(cblk->mqdec = mqdec_create(NUMCTXS, 0))) {
					return -1;
				}
				mqdec_setctxs(cblk->mqdec, NUMCTXS, mqctxs);
			}
			mqdec_setinput(cblk->mqdec, seg->stream);
			mqdec_init(cblk->mqdec);
		} else {
			assert(seg->type == SEG_RAW);
			if (!cblk->nulldec) {
				if (!(cblk->nulldec = bitstream_sopen(seg->stream, "r"))) {
					assert(0);
				}
			}
		}


		for (i = 0; i < seg->numpasses; ++i) {
			bpno = band->roishift + band->numbps - 1 - (cblk->numimsbs +
			  (seg->passno + i - cblk->firstpassno + 2) / 3);
#if 1
			passtype = (seg->passno + i + 2) % 3;
#else
			passtype = PASSTYPE(seg->passno + i + 2);
#endif
			assert(bpno >= 0 && bpno < 32);
			switch (passtype) {
			case SIGPASS:
				ret = (seg->type == SEG_MQ) ? dec_sigpass(dec,
				  cblk->mqdec, bpno, band->orient,
				  (tile->cp->ccps[compno].cblkctx & JPC_COX_VSC) != 0,
				  cblk->flags, cblk->data) :
				  dec_rawsigpass(dec, cblk->nulldec, bpno,
				  (tile->cp->ccps[compno].cblkctx & JPC_COX_VSC) != 0,
				  cblk->flags, cblk->data);
				break;
			case REFPASS:
				ret = (seg->type == SEG_MQ) ?
				  dec_refpass(dec, cblk->mqdec, bpno,
				  (tile->cp->ccps[compno].cblkctx & JPC_COX_VSC) != 0,
				  cblk->flags, cblk->data) :
				  dec_rawrefpass(dec, cblk->nulldec, bpno,
				  (tile->cp->ccps[compno].cblkctx & JPC_COX_VSC) != 0,
				  cblk->flags, cblk->data);
				break;
			case CLNPASS:
				assert(seg->type == SEG_MQ);
				ret = dec_clnpass(dec, cblk->mqdec, bpno,
				  band->orient, (tile->cp->ccps[compno].cblkctx &
				  JPC_COX_VSC) != 0, (tile->cp->ccps[compno].cblkctx &
				  JPC_COX_SEGSYM) != 0, cblk->flags,
				  cblk->data);
				break;
			default:
				ret = -1;
				break;
			}
			/* Do we need to reset after each coding pass? */
			if (tile->cp->ccps[compno].cblkctx & JPC_COX_RESET) {
				mqdec_setctxs(cblk->mqdec, NUMCTXS, mqctxs);
			}

			if (ret) {
				fprintf(stderr, "coding pass failed passtype=%d segtype=%d\n", passtype, seg->type);
				return -1;
			}

		}

		if (seg->type == SEG_MQ) {
/* Note: dont destroy mq decoder because context info will be lost */
		} else {
			assert(seg->type == SEG_RAW);
			if (tile->cp->ccps[compno].cblkctx & JPC_COX_PTERM) {
				fillmask = 0x7f;
				filldata = 0x2a;
			} else {
				fillmask = 0;
				filldata = 0;
			}
			if ((ret = bitstream_inalign(cblk->nulldec, fillmask,
			  filldata)) < 0) {
				return -1;
			} else if (ret > 0) {
				fprintf(stderr, "warning: bad termination pattern detected\n");
			}
			bitstream_close(cblk->nulldec);
			cblk->nulldec = 0;
		}

		cblk->curseg = seg->next;
		seglist_remove(&cblk->segs, seg);
		seg_destroy(seg);
		seg = cblk->curseg;
	}

	assert(dopartial ? (!cblk->curseg) : 1);

	return 0;
}

/******************************************************************************\
* Code for significance pass.
\******************************************************************************/

#define	sigpass_step(fp, frowstep, dp, bitpos, oneplushalf, orient, mqdec, vcausalflag) \
{ \
	int f; \
	int v; \
	f = *(fp); \
	if ((f & OTHSIGMSK) && !(f & (SIG | VISIT))) { \
		mqdec_setcurctx((mqdec), GETZCCTXNO(f, (orient))); \
		T1D_GETBIT((mqdec), v, "SIG", "ZC"); \
		if (v) { \
			mqdec_setcurctx((mqdec), GETSCCTXNO(f)); \
			T1D_GETBIT((mqdec), v, "SIG", "SC"); \
			v ^= GETSPB(f); \
			UPDATEFLAGS4((fp), (frowstep), v, (vcausalflag)); \
			*(fp) |= SIG; \
			*(dp) = (v) ? (-(oneplushalf)) : (oneplushalf); \
		} \
		*(fp) |= VISIT; \
	} \
}

static int dec_sigpass(jpd_t *dec, register mqdec_t *mqdec, int bitpos, int orient,
  int vcausalflag, matrix_t *flags, matrix_t *data)
{
	int i;
	int j;
	int one;
	int half;
	int oneplushalf;
	int vscanlen;
	int width;
	int height;
	fix_t *fp;
	int frowstep;
	int fstripestep;
	fix_t *fstripestart;
	fix_t *fvscanstart;
	fix_t *dp;
	int drowstep;
	int dstripestep;
	fix_t *dstripestart;
	fix_t *dvscanstart;
	int k;

	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << bitpos;
	half = one >> 1;
	oneplushalf = one | half;

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			/* Process first sample in vertical scan. */
			sigpass_step(fp, frowstep, dp, bitpos, oneplushalf,
			  orient, mqdec, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process second sample in vertical scan. */
			sigpass_step(fp, frowstep, dp, bitpos, oneplushalf,
			  orient, mqdec, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process third sample in vertical scan. */
			sigpass_step(fp, frowstep, dp, bitpos, oneplushalf,
			  orient, mqdec, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process fourth sample in vertical scan. */
			sigpass_step(fp, frowstep, dp, bitpos, oneplushalf,
			  orient, mqdec, 0);
		}
	}
	return 0;
}

#define	rawsigpass_step(fp, frowstep, dp, oneplushalf, in, vcausalflag) \
{ \
	fix_t f = *(fp); \
	fix_t v; \
	if ((f & OTHSIGMSK) && !(f & (SIG | VISIT))) { \
		T1D_RAWGETBIT(in, v, "SIG", "ZC"); \
		if (v < 0) { \
			return -1; \
		} \
		if (v) { \
			T1D_RAWGETBIT(in, v, "SIG", "SC"); \
			if (v < 0) { \
				return -1; \
			} \
			UPDATEFLAGS4((fp), (frowstep), v, (vcausalflag)); \
			*(fp) |= SIG; \
			*(dp) = v ? (-oneplushalf) : (oneplushalf); \
		} \
		*(fp) |= VISIT; \
	} \
}

static int dec_rawsigpass(jpd_t *dec, bitstream_t *in, int bitpos, int vcausalflag,
  matrix_t *flags, matrix_t *data)
{
	int i;
	int j;
	int k;
	int one;
	int half;
	int oneplushalf;
	int vscanlen;
	int width;
	int height;
	fix_t *fp;
	int frowstep;
	int fstripestep;
	fix_t *fstripestart;
	fix_t *fvscanstart;
	fix_t *dp;
	int drowstep;
	int dstripestep;
	fix_t *dstripestart;
	fix_t *dvscanstart;

	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << bitpos;
	half = one >> 1;
	oneplushalf = one | half;

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			/* Process first sample in vertical scan. */
			rawsigpass_step(fp, frowstep, dp, oneplushalf,
			  in, vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process second sample in vertical scan. */
			rawsigpass_step(fp, frowstep, dp, oneplushalf,
			  in, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process third sample in vertical scan. */
			rawsigpass_step(fp, frowstep, dp, oneplushalf,
			  in, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process fourth sample in vertical scan. */
			rawsigpass_step(fp, frowstep, dp, oneplushalf,
			  in, 0);

		}
	}
	return 0;
}

/******************************************************************************\
* Code for refinement pass.
\******************************************************************************/

#define	refpass_step(fp, dp, poshalf, neghalf, mqdec, vcausalflag) \
{ \
	int v; \
	int t; \
	if (((*(fp)) & (SIG | VISIT)) == SIG) { \
		mqdec_setcurctx((mqdec), GETMAGCTXNO(*(fp))); \
		T1D_GETBITNOSKEW((mqdec), v, "REF", "MR"); \
		t = (v ? (poshalf) : (neghalf)); \
		*(dp) += (*(dp) < 0) ? (-t) : t; \
		*(fp) |= REFINE; \
	} \
}

static int dec_refpass(jpd_t *dec, register mqdec_t *mqdec, int bitpos,
  int vcausalflag, matrix_t *flags, matrix_t *data)
{
	int i;
	int j;
	int vscanlen;
	int width;
	int height;
	int one;
	int poshalf;
	int neghalf;
	fix_t *fp;
	int frowstep;
	int fstripestep;
	fix_t *fstripestart;
	fix_t *fvscanstart;
	fix_t *dp;
	int drowstep;
	int dstripestep;
	fix_t *dstripestart;
	fix_t *dvscanstart;
	int k;

	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << bitpos;
	poshalf = one >> 1;
	neghalf = (bitpos > 0) ? (-poshalf) : (-1);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			/* Process first sample in vertical scan. */
			refpass_step(fp, dp, poshalf, neghalf, mqdec,
			  vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process second sample in vertical scan. */
			refpass_step(fp, dp, poshalf, neghalf, mqdec, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process third sample in vertical scan. */
			refpass_step(fp, dp, poshalf, neghalf, mqdec, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process fourth sample in vertical scan. */
			refpass_step(fp, dp, poshalf, neghalf, mqdec, 0);
		}
	}

	return 0;
}

#define	rawrefpass_step(fp, dp, poshalf, neghalf, in, vcausalflag) \
{ \
	fix_t v; \
	fix_t t; \
	if (((*(fp)) & (SIG | VISIT)) == SIG) { \
		T1D_RAWGETBIT(in, v, "REF", "MAGREF"); \
		if (v < 0) { \
			return -1; \
		} \
		t = (v ? poshalf : neghalf); \
		*(dp) += (*(dp) < 0) ? (-t) : t; \
		*(fp) |= REFINE; \
	} \
}

static int dec_rawrefpass(jpd_t *dec, bitstream_t *in, int bitpos, int vcausalflag,
  matrix_t *flags, matrix_t *data)
{
	int i;
	int j;
	int k;
	int vscanlen;
	int width;
	int height;
	int one;
	int poshalf;
	int neghalf;
	fix_t *fp;
	int frowstep;
	int fstripestep;
	fix_t *fstripestart;
	fix_t *fvscanstart;
	fix_t *dp;
	int drowstep;
	int dstripestep;
	fix_t *dstripestart;
	fix_t *dvscanstart;

	width = matrix_numcols(data);
	height = matrix_numrows(data);
	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	one = 1 << bitpos;
	poshalf = one >> 1;
	neghalf = (bitpos > 0) ? (-poshalf) : (-1);

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = height; i > 0; i -= 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(i, 4);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			dp = dvscanstart;
			k = vscanlen;

			/* Process first sample in vertical scan. */
			rawrefpass_step(fp, dp, poshalf, neghalf, in,
			  vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process second sample in vertical scan. */
			rawrefpass_step(fp, dp, poshalf, neghalf, in, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process third sample in vertical scan. */
			rawrefpass_step(fp, dp, poshalf, neghalf, in, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process fourth sample in vertical scan. */
			rawrefpass_step(fp, dp, poshalf, neghalf, in, 0);
		}
	}
	return 0;
}

/******************************************************************************\
* Code for cleanup pass.
\******************************************************************************/

#define	clnpass_step(f, fp, frowstep, dp, oneplushalf, orient, mqdec, flabel, plabel, vcausalflag) \
{ \
	int v; \
flabel \
	if (!((f) & (SIG | VISIT))) { \
		mqdec_setcurctx((mqdec), GETZCCTXNO((f), (orient))); \
		T1D_GETBIT((mqdec), v, "CLN", "ZC"); \
		if (v) { \
plabel \
			/* Coefficient is significant. */ \
			mqdec_setcurctx((mqdec), GETSCCTXNO(f)); \
			T1D_GETBIT((mqdec), v, "CLN", "SC"); \
			v ^= GETSPB(f); \
			*(dp) = (v) ? (-(oneplushalf)) : (oneplushalf); \
			UPDATEFLAGS4((fp), (frowstep), v, (vcausalflag)); \
			*(fp) |= SIG; \
		} \
	} \
	/* XXX - Is this correct?  Can aggregation cause some VISIT bits not to be reset?  Check. */ \
	*(fp) &= ~VISIT; \
}

static int dec_clnpass(jpd_t *dec, register mqdec_t *mqdec, int bitpos, int orient,
  int vcausalflag, int segsymflag, matrix_t *flags, matrix_t *data)
{
	int i;
	int j;
	int k;
	int vscanlen;
	int v;
	int half;
	int runlen;
	int f;
	int width;
	int height;
	int one;
	int oneplushalf;

	fix_t *fp;
	int frowstep;
	int fstripestep;
	fix_t *fstripestart;
	fix_t *fvscanstart;

	fix_t *dp;
	int drowstep;
	int dstripestep;
	fix_t *dstripestart;
	fix_t *dvscanstart;

	one = 1 << bitpos;
	half = one >> 1;
	oneplushalf = one | half;

	width = matrix_numcols(data);
	height = matrix_numrows(data);

	frowstep = matrix_rowstep(flags);
	drowstep = matrix_rowstep(data);
	fstripestep = frowstep << 2;
	dstripestep = drowstep << 2;

	fstripestart = matrix_getref(flags, 1, 1);
	dstripestart = matrix_getref(data, 0, 0);
	for (i = 0; i < height; i += 4, fstripestart += fstripestep,
	  dstripestart += dstripestep) {
		fvscanstart = fstripestart;
		dvscanstart = dstripestart;
		vscanlen = min(4, height - i);
		for (j = width; j > 0; --j, ++fvscanstart, ++dvscanstart) {
			fp = fvscanstart;
			if (vscanlen >= 4 && (!((*fp) & (SIG | VISIT |
			  OTHSIGMSK))) && (fp += frowstep, !((*fp) & (SIG |
			  VISIT | OTHSIGMSK))) && (fp += frowstep, !((*fp) &
			  (SIG | VISIT | OTHSIGMSK))) && (fp += frowstep,
			  !((*fp) & (SIG | VISIT | OTHSIGMSK)))) {

				mqdec_setcurctx(mqdec, AGGCTXNO);
				T1D_GETBIT(mqdec, v, "CLN", "AGG");
				if (!v) {
					continue;
				}
				mqdec_setcurctx(mqdec, UCTXNO);
				T1D_GETBITNOSKEW(mqdec, v, "CLN", "RL");
				runlen = v;
				T1D_GETBITNOSKEW(mqdec, v, "CLN", "RL");
				runlen = (runlen << 1) | v;
				f = *(fp = fvscanstart + frowstep * runlen);
				dp = dvscanstart + drowstep * runlen;
				k = vscanlen - runlen;
				switch (runlen) {
				case 0:
					goto clnpass_partial0;
					break;
				case 1:
					goto clnpass_partial1;
					break;
				case 2:
					goto clnpass_partial2;
					break;
				case 3:
					goto clnpass_partial3;
					break;
				}
			} else {
				f = *(fp = fvscanstart);
				dp = dvscanstart;
				k = vscanlen;
				goto clnpass_full0;
			}

			/* Process first sample in vertical scan. */
			clnpass_step(f, fp, frowstep, dp, oneplushalf, orient,
			  mqdec, clnpass_full0:, clnpass_partial0:,
			  vcausalflag);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process second sample in vertical scan. */
			f = *fp;
			clnpass_step(f, fp, frowstep, dp, oneplushalf, orient,
				mqdec, ;, clnpass_partial1:, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process third sample in vertical scan. */
			f = *fp;
			clnpass_step(f, fp, frowstep, dp, oneplushalf, orient,
				mqdec, ;, clnpass_partial2:, 0);
			if (--k <= 0) {
				continue;
			}
			fp += frowstep;
			dp += drowstep;

			/* Process fourth sample in vertical scan. */
			f = *fp;
			clnpass_step(f, fp, frowstep, dp, oneplushalf, orient,
				mqdec, ;, clnpass_partial3:, 0);
		}
	}

	if (segsymflag) {
		int segsymval;
		segsymval = 0;
		mqdec_setcurctx(mqdec, UCTXNO);
		T1D_GETBITNOSKEW(mqdec, v, "CLN", "SEGSYM");
		segsymval = (segsymval << 1) | (v & 1);
		T1D_GETBITNOSKEW(mqdec, v, "CLN", "SEGSYM");
		segsymval = (segsymval << 1) | (v & 1);
		T1D_GETBITNOSKEW(mqdec, v, "CLN", "SEGSYM");
		segsymval = (segsymval << 1) | (v & 1);
		T1D_GETBITNOSKEW(mqdec, v, "CLN", "SEGSYM");
		segsymval = (segsymval << 1) | (v & 1);
		if (segsymval != 0xa) {
			fprintf(stderr, "warning: bad segmentation symbol\n");
		}
	}

	return 0;
}
