/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Tree-Structured Filter Bank (TSFB) Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <assert.h>
#include "tsfb.h"
#include "seq.h"
#include "jpccod.h"
#include "memalloc.h"

/******************************************************************************\
*
\******************************************************************************/

#define	bandnotovind(tsfbnode, x)	((x) / (tsfbnode)->numhchans)
#define	bandnotohind(tsfbnode, x)	((x) % (tsfbnode)->numhchans)

static tsfb_t *tsfb_create();
static tsfbnode_t *tsfbnode_create();
static void tsfbnode_destroy(tsfbnode_t *node);
static void tsfbnode_synthesize(tsfbnode_t *node, int flags, seq2d_t *x);
static void tsfbnode_analyze(tsfbnode_t *node, int flags, seq2d_t *x);
static void qmfb2d_getbands(qmfb1d_t *hqmfb, qmfb1d_t *vqmfb,
  uint_fast32_t xstart, uint_fast32_t ystart, uint_fast32_t xend,
  uint_fast32_t yend, int maxbands, int *numbandsptr, tsfbnodeband_t *bands);
static void tsfbnode_getbandstree(tsfbnode_t *node, uint_fast32_t posxstart,
  uint_fast32_t posystart, uint_fast32_t xstart, uint_fast32_t ystart,
  uint_fast32_t xend, uint_fast32_t yend, tsfb_band_t **bands);
static int tsfbnode_findchild(tsfbnode_t *parnode, tsfbnode_t *cldnode);
static int tsfbnode_getequivfilters(tsfbnode_t *tsfbnode, int cldind,
  int width, int height, seq_t **vfilter, seq_t **hfilter);

/******************************************************************************\
*
\******************************************************************************/

tsfb_t *tsfb_wavelet(qmfb1d_t *hqmfb, qmfb1d_t *vqmfb, int numdlvls)
{
	tsfb_t *tsfb;
	int dlvlno;
	tsfbnode_t *curnode;
	tsfbnode_t *prevnode;
	int childno;
	if (!(tsfb = tsfb_create())) {
		return 0;
	}
	prevnode = 0;
	for (dlvlno = 0; dlvlno < numdlvls; ++dlvlno) {
		if (!(curnode = tsfbnode_create())) {
			tsfb_destroy(tsfb);
			return 0;
		}
		if (prevnode) {
			prevnode->children[0] = curnode;
			++prevnode->numchildren;
			curnode->parent = prevnode;
		} else {
			tsfb->root = curnode;
			curnode->parent = 0;
		}
		if (hqmfb) {
			curnode->numhchans = qmfb1d_getnumchans(hqmfb);
			if (!(curnode->hqmfb = qmfb1d_copy(hqmfb))) {
				tsfb_destroy(tsfb);
				return 0;
			}
		} else {
			curnode->hqmfb = 0;
			curnode->numhchans = 1;
		}
		if (vqmfb) {
			curnode->numvchans = qmfb1d_getnumchans(vqmfb);
			if (!(curnode->vqmfb = qmfb1d_copy(vqmfb))) {
				tsfb_destroy(tsfb);
				return 0;
			}
		} else {
			curnode->vqmfb = 0;
			curnode->numvchans = 1;
		}
		curnode->maxchildren = curnode->numhchans * curnode->numvchans;
		for (childno = 0; childno < curnode->maxchildren;
		  ++childno) {
			curnode->children[childno] = 0;
		}
		prevnode = curnode;
	}
	return tsfb;
}

static tsfb_t *tsfb_create()
{
	tsfb_t *tsfb;
	if (!(tsfb = malloc(sizeof(tsfb_t)))) {
		return 0;
	}
	tsfb->root = 0;
	return tsfb;
}

void tsfb_destroy(tsfb_t *tsfb)
{
	if (!tsfb->root)
		return;
	tsfbnode_destroy(tsfb->root);
	free(tsfb);
}

/******************************************************************************\
*
\******************************************************************************/

void tsfb_analyze(tsfb_t *tsfb, int flags, seq2d_t *x)
{
	if (tsfb->root) {
		tsfbnode_analyze(tsfb->root, flags, x);
	}
}

static void tsfbnode_analyze(tsfbnode_t *node, int flags, seq2d_t *x)
{
	tsfbnodeband_t nodebands[TSFB_MAXBANDSPERNODE];
	int numbands;
	seq2d_t *y;
	int bandno;
	tsfbnodeband_t *band;

	if (node->vqmfb) {
		qmfb1d_analyze(node->vqmfb, flags | QMFB1D_VERT, x);
	}
	if (node->hqmfb) {
		qmfb1d_analyze(node->hqmfb, flags, x);
	}
	if (node->numchildren > 0) {
		qmfb2d_getbands(node->hqmfb, node->vqmfb, seq2d_xstart(x),
		  seq2d_ystart(x), seq2d_xend(x), seq2d_yend(x),
		  TSFB_MAXBANDSPERNODE, &numbands, nodebands);
		y = seq2d_create(0, 0, 0, 0);
		assert(y);
		for (bandno = 0, band = nodebands; bandno < numbands; ++bandno, ++band) {
			if (node->children[bandno]) {
				if (band->xstart != band->xend && band->ystart != band->yend) {
					seq2d_bindsub(y, x, band->locxstart, band->locystart,
					  band->locxend, band->locyend);
					seq2d_setshift(y, band->xstart, band->ystart);
					tsfbnode_analyze(node->children[bandno], flags, y);
				}
			}
		}
		matrix_destroy(y);
	}
}

void tsfb_synthesize(tsfb_t *tsfb, int flags, seq2d_t *x)
{
	if (tsfb->root) {
		tsfbnode_synthesize(tsfb->root, flags, x);
	}
}

static void tsfbnode_synthesize(tsfbnode_t *node, int flags, seq2d_t *x)
{
	tsfbnodeband_t nodebands[TSFB_MAXBANDSPERNODE];
	int numbands;
	seq2d_t *y;
	int bandno;
	tsfbnodeband_t *band;

	if (node->numchildren > 0) {
		qmfb2d_getbands(node->hqmfb, node->vqmfb, seq2d_xstart(x),
		  seq2d_ystart(x), seq2d_xend(x), seq2d_yend(x),
		  TSFB_MAXBANDSPERNODE, &numbands, nodebands);
		y = seq2d_create(0, 0, 0, 0);
		for (bandno = 0, band = nodebands; bandno < numbands; ++bandno, ++band) {
			if (node->children[bandno]) {
				if (band->xstart != band->xend && band->ystart != band->yend) {
					seq2d_bindsub(y, x, band->locxstart, band->locystart,
					  band->locxend, band->locyend);
					seq2d_setshift(y, band->xstart, band->ystart);
					tsfbnode_synthesize(node->children[bandno], flags, y);
				}
			}
		}
		seq2d_destroy(y);
	}
	if (node->hqmfb) {
		qmfb1d_synthesize(node->hqmfb, flags, x);
	}
	if (node->vqmfb) {
		qmfb1d_synthesize(node->vqmfb, flags | QMFB1D_VERT, x);
	}
}

/******************************************************************************\
*
\******************************************************************************/


int tsfb_getbands(tsfb_t *tsfb, uint_fast32_t xstart, uint_fast32_t ystart,
  uint_fast32_t xend, uint_fast32_t yend, tsfb_band_t *bands)
{
	tsfb_band_t *savbands;
	savbands = bands;
	if (!tsfb->root) {
		bands[0].xstart = xstart;
		bands[0].ystart = ystart;
		bands[0].xend = xend;
		bands[0].yend = yend;
		bands[0].locxstart = xstart;
		bands[0].locystart = ystart;
		bands[0].locxend = xend;
		bands[0].locyend = yend;
		bands[0].orient = JP_LL;
		bands[0].synenergywt = FIX_ONE;
		++bands;
	} else {
		tsfbnode_getbandstree(tsfb->root, xstart, ystart,
		  xstart, ystart, xend, yend, &bands);
	}
	return bands - savbands;
}

static void tsfbnode_getbandstree(tsfbnode_t *node, uint_fast32_t posxstart,
  uint_fast32_t posystart, uint_fast32_t xstart, uint_fast32_t ystart,
  uint_fast32_t xend, uint_fast32_t yend, tsfb_band_t **bands)
{
	tsfbnodeband_t nodebands[TSFB_MAXBANDSPERNODE];
	tsfbnodeband_t *nodeband;
	int nodebandno;
	int numnodebands;
	tsfb_band_t *band;
	seq_t *hfilter;
	seq_t *vfilter;

	qmfb2d_getbands(node->hqmfb, node->vqmfb, xstart, ystart, xend, yend,
	  TSFB_MAXBANDSPERNODE, &numnodebands, nodebands);
	if (node->numchildren > 0) {
		for (nodebandno = 0, nodeband = nodebands;
		  nodebandno < numnodebands; ++nodebandno, ++nodeband) {
			if (node->children[nodebandno]) {
				tsfbnode_getbandstree(node->children[
				  nodebandno], posxstart +
				  nodeband->locxstart - xstart, posystart +
				  nodeband->locystart - ystart, nodeband->xstart,
				  nodeband->ystart, nodeband->xend,
				  nodeband->yend, bands);

			}
		}
	}
assert(numnodebands == 4 || numnodebands == 3);
	for (nodebandno = 0, nodeband = nodebands; nodebandno < numnodebands;
	  ++nodebandno, ++nodeband) {
		if (!node->children[nodebandno]) {
			band = *bands;
			band->xstart = nodeband->xstart;
			band->ystart = nodeband->ystart;
			band->xend = nodeband->xend;
			band->yend = nodeband->yend;
			band->locxstart = posxstart + nodeband->locxstart -
			  xstart;
			band->locystart = posystart + nodeband->locystart -
			  ystart;
			band->locxend = band->locxstart + band->xend -
			  band->xstart;
			band->locyend = band->locystart + band->yend -
			  band->ystart;
			if (numnodebands == 4) {
				switch (nodebandno) {
				case 0:
					band->orient = JP_LL;
					break;
				case 1:
					band->orient = JP_HL;
					break;
				case 2:
					band->orient = JP_LH;
					break;
				case 3:
					band->orient = JP_HH;
					break;
				default:
					abort();
					break;
				}
			} else {
				switch (nodebandno) {
				case 0:
					band->orient = JP_HL;
					break;
				case 1:
					band->orient = JP_LH;
					break;
				case 2:
					band->orient = JP_HH;
					break;
				default:
					abort();
					break;
				}
			}
			tsfbnode_getequivfilters(node, nodebandno, band->xend - band->xstart, band->yend - band->ystart, &hfilter, &vfilter);
			band->synenergywt = fix_mul(seq_norm(hfilter),
			  seq_norm(vfilter));
			seq_destroy(hfilter);
			seq_destroy(vfilter);
			++(*bands);
		}
	}
}

/******************************************************************************\
*
\******************************************************************************/

void tsfb_prune(tsfb_t *tsfb, int flags, int maxlvls, seq2d_t *x)
{
	abort();
}

int tsfb_getmaxlvls(tsfb_t *tsfb)
{
	abort();
	return 0;
}

/******************************************************************************\
*
\******************************************************************************/

static tsfbnode_t *tsfbnode_create()
{
	tsfbnode_t *node;
	if (!(node = malloc(sizeof(tsfbnode_t)))) {
		return 0;
	}
	node->numhchans = 0;
	node->numvchans = 0;
	node->numchildren = 0;
	node->maxchildren = 0;
	node->hqmfb = 0;
	node->vqmfb = 0;
	node->parent = 0;
	return node;
}

static void tsfbnode_destroy(tsfbnode_t *node)
{
	tsfbnode_t **child;
	int childno;
	for (childno = 0, child = node->children; childno < node->maxchildren;
	  ++childno, ++child) {
		if (*child) {
			tsfbnode_destroy(*child);
		}
	}
	if (node->hqmfb) {
		qmfb1d_destroy(node->hqmfb);
	}
	if (node->vqmfb) {
		qmfb1d_destroy(node->vqmfb);
	}
	free(node);
}








static void qmfb2d_getbands(qmfb1d_t *hqmfb, qmfb1d_t *vqmfb,
  uint_fast32_t xstart, uint_fast32_t ystart, uint_fast32_t xend,
  uint_fast32_t yend, int maxbands, int *numbandsptr, tsfbnodeband_t *bands)
{
	qmfb1dband_t hbands[QMFB1D_MAXCHANS];
	qmfb1dband_t vbands[QMFB1D_MAXCHANS];
	int numhbands;
	int numvbands;
	int numbands;
	int bandno;
	int hbandno;
	int vbandno;
	tsfbnodeband_t *band;

	if (hqmfb) {
		qmfb1d_getbands(hqmfb, 0, xstart, ystart, xend, yend,
		  QMFB1D_MAXCHANS, &numhbands, hbands);
	} else {
		numhbands = 1;
		hbands[0].start = xstart;
		hbands[0].end = xend;
		hbands[0].locstart = xstart;
		hbands[0].locend = xend;
	}
	if (vqmfb) {
		qmfb1d_getbands(vqmfb, QMFB1D_VERT, xstart, ystart, xend,
		  yend, QMFB1D_MAXCHANS, &numvbands, vbands);
	} else {
		numvbands = 1;
		vbands[0].start = ystart;
		vbands[0].end = yend;
		vbands[0].locstart = ystart;
		vbands[0].locend = yend;
	}
	numbands = numhbands * numvbands;
	*numbandsptr = numbands;
	for (bandno = 0, band = bands; bandno < numbands; ++bandno, ++band) {
		hbandno = bandno % numhbands;
		vbandno = bandno / numhbands;
		band->xstart = hbands[hbandno].start;
		band->ystart = vbands[vbandno].start;
		band->xend = hbands[hbandno].end;
		band->yend = vbands[vbandno].end;
		band->locxstart = hbands[hbandno].locstart;
		band->locystart = vbands[vbandno].locstart;
		band->locxend = hbands[hbandno].locend;
		band->locyend = vbands[vbandno].locend;
		assert(band->xstart <= band->xend &&
		  band->ystart <= band->yend);
		if (band->xstart == band->xend) {
			band->yend = band->ystart;
			band->locyend = band->locystart;
		} else if (band->ystart == band->yend) {
			band->xend = band->xstart;
			band->locxend = band->locxstart;
		}
	}
}

static int tsfbnode_getequivfilters(tsfbnode_t *tsfbnode, int cldind,
  int width, int height, seq_t **hfilter, seq_t **vfilter)
{
	seq_t *hseq;
	seq_t *vseq;
	tsfbnode_t *node;
	seq2d_t *hfilters[QMFB1D_MAXCHANS];
	seq2d_t *vfilters[QMFB1D_MAXCHANS];
	int numhchans;
	int numvchans;
	seq_t *tmpseq;

	hseq = 0;
	vseq = 0;

	if (!(hseq = seq_create(0, 1))) {
		goto error;
	}
	seq_set(hseq, 0, inttofix(1));
	if (!(vseq = seq_create(0, 1))) {
		goto error;
	}
	seq_set(vseq, 0, inttofix(1));

	node = tsfbnode;
	while (node) {
		if (node->hqmfb) {
			numhchans = qmfb1d_getnumchans(node->hqmfb);
			if (qmfb1d_getsynfilters(node->hqmfb, width, hfilters)) {
				goto error;
			}
			if (!(tmpseq = seq_upsample(hseq, numhchans))) {
				goto error;
			}
			seq_destroy(hseq);
			hseq = tmpseq;
			if (!(tmpseq = seq_conv(hseq, hfilters[bandnotohind(node, cldind)]))) {
				goto error;
			}
			seq_destroy(hfilters[0]);
			seq_destroy(hfilters[1]);
			seq_destroy(hseq);
			hseq = tmpseq;
		}
		if (node->vqmfb) {
			numvchans = qmfb1d_getnumchans(node->vqmfb);
			if (qmfb1d_getsynfilters(node->vqmfb, height, vfilters)) {
				abort();
			}
			if (!(tmpseq = seq_upsample(vseq, numvchans))) {
				goto error;
			}
			seq_destroy(vseq);
			vseq = tmpseq;
			if (!(tmpseq = seq_conv(vseq, vfilters[bandnotovind(node, cldind)]))) {
				goto error;
			}
			seq_destroy(vfilters[0]);
			seq_destroy(vfilters[1]);
			seq_destroy(vseq);
			vseq = tmpseq;
		}
		if (node->parent) {
			cldind = tsfbnode_findchild(node->parent, node);
		}
		node = node->parent;
	}

	*hfilter = hseq;
	*vfilter = vseq;

	return 0;

error:
	if (hseq) {
		seq_destroy(hseq);
	}
	if (vseq) {
		seq_destroy(vseq);
	}
	return -1;

}

static int tsfbnode_findchild(tsfbnode_t *parnode, tsfbnode_t *cldnode)
{
	int i;

	for (i = 0; i < parnode->maxchildren; i++) {
		if (parnode->children[i] == cldnode)
			return i;
	}
	assert(0);
	return -1;
}
